from __future__ import annotations

from enum import Enum, auto
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Iterator, cast
from logging import getLogger

from app_model.types import MenuItem
from app_model.backends.qt import QCommandRuleAction
from qtpy import QtWidgets as QtW
from qtpy import QtCore, QtGui
from qtpy.QtCore import Qt
from superqt import QIconifyIcon
from superqt.utils import qthrottled

from himena.workflow import LocalReaderMethod
from himena._utils import get_display_name
from himena.utils.misc import lru_cache
from himena.consts import MenuId
from himena.types import DragDataModel, WindowState, WindowRect, Size
from himena.plugins import _checker
from himena.utils.window_rect import ResizeState
from himena.qt._utils import get_main_window, build_qmodel_menu
from himena.qt._qrename import QRenameLineEdit
from himena import _drag

if TYPE_CHECKING:
    from PyQt6 import QtWidgets as QtW
    from himena.qt.main_window import MainWindowQt
    from himena.widgets import SubWindow

_LOGGER = getLogger(__name__)


class QSubWindowArea(QtW.QMdiArea):
    area_focused = QtCore.Signal()

    def __init__(self):
        super().__init__()
        self.viewport().setFocusPolicy(Qt.FocusPolicy.ClickFocus)
        self._last_press_pos: QtCore.QPoint | None = None
        self._last_drag_pos: QtCore.QPoint | None = None
        self.setActivationOrder(QtW.QMdiArea.WindowOrder.ActivationHistoryOrder)
        self._last_active_window_id: QSubWindow | None = None

    def addSubWindow(self, sub_window: QSubWindow):
        super().addSubWindow(sub_window)
        sub_window.show()

    def iter_widgets(self) -> Iterator[QtW.QWidget]:
        """Iterate over all widgets in the sub-window area."""
        for sub_window in self.subWindowList():
            yield sub_window.widget()

    def indexOf(self, sub_window: QSubWindow) -> int:
        return self.subWindowList().index(sub_window)

    def relabel_widgets(self):
        """Update the 0, 1, 2... labels in the sub-windows."""
        for i, sub_window in enumerate(self.subWindowList()):
            sub_window._title_bar._index_label.setText(str(i))

    def add_widget(
        self,
        widget: QtW.QWidget,
        title: str | None = None,
    ) -> QSubWindow:
        if title is None:
            title = widget.objectName() or "Window"
        if not isinstance(widget, QtW.QWidget):
            raise TypeError(
                f"`widget` must be a QtW.QWidget instance, got {type(widget)}."
            )
        size = widget.sizeHint().expandedTo(QtCore.QSize(160, 120)) + QtCore.QSize(8, 8)
        sub_window = QSubWindow(widget, title)
        self.addSubWindow(sub_window)
        sub_window.resize(size)
        self.relabel_widgets()
        return sub_window

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        self._last_drag_pos = self._last_press_pos = event.pos()
        if (
            event.buttons() & Qt.MouseButton.LeftButton
            and event.modifiers() & Qt.KeyboardModifier.ControlModifier
        ):
            self.setCursor(Qt.CursorShape.ClosedHandCursor)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        self._mouse_move_event(event)

    def _mouse_move_event(self, event: QtGui.QMouseEvent):
        if event.buttons() & Qt.MouseButton.LeftButton:
            if self._last_drag_pos is None:
                return None
            if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
                # move all the windows
                dpos = event.pos() - self._last_drag_pos
                for sub_window in self.subWindowList():
                    sub_window.move(sub_window.pos() + dpos)
                self.setCursor(Qt.CursorShape.ClosedHandCursor)
        self._last_drag_pos = event.pos()

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        # reset cursor state
        self.setCursor(Qt.CursorShape.ArrowCursor)
        self._last_press_pos = self._last_drag_pos = None

        # check if any window is under the cursor
        for sub_window in self.subWindowList():
            if sub_window.rect().contains(sub_window.mapFromParent(event.pos())):
                if not sub_window.is_current():
                    sub_window.set_is_current(True)
                    self.subWindowActivated.emit(sub_window)
                break
        else:
            if event.button() == Qt.MouseButton.RightButton:
                # context menu
                app = get_main_window(self).model_app
                menu = build_qmodel_menu(MenuId.FILE_NEW, app, self)
                menu.exec(event.globalPos())

    def eventFilter(self, obj, a0: QtCore.QEvent) -> bool:
        if a0.type() == QtCore.QEvent.Type.FocusIn:
            with suppress(RuntimeError):
                if obj is not self:
                    if isinstance(obj, (QtW.QStyle, QtW.QAbstractButton, QtW.QMenuBar)):
                        return False
                    else:
                        if self._last_active_window_id is None and (
                            win := self.currentSubWindow()
                        ):
                            self.subWindowActivated.emit(win)
                            self._last_active_window_id = win._my_wrapper()._identifier
                            _LOGGER.debug("QSubWindowArea.eventFilter: Window focused.")
                else:
                    self._set_area_focused()
                    _LOGGER.debug("QSubWindowArea.eventFilter: TabArea focused.")
        try:
            return super().eventFilter(obj, a0)
        except RuntimeError:  # wrapped object deleted
            return False

    def _set_area_focused(self):
        self.area_focused.emit()
        self._last_active_window_id = None

    def hideEvent(self, a0: QtGui.QHideEvent | None) -> None:
        self._last_drag_pos = self._last_press_pos = None
        return super().hideEvent(a0)

    def _pixmap_resized(
        self,
        size: QtCore.QSize,
        outline: QtGui.QColor | None = None,
    ) -> QtGui.QPixmap:
        return pixmap_resized(self, size, outline)

    if TYPE_CHECKING:

        def subWindowList(self) -> list[QSubWindow]: ...
        def activeSubWindow(self) -> QSubWindow: ...
        def currentSubWindow(self) -> QSubWindow | None: ...


def _get_icon(name: str, rotate=None, color=None):
    try:
        return QIconifyIcon(name, rotate=rotate, color=color)
    except OSError:
        return QtGui.QIcon()


class TitleIconId(Enum):
    WINDOW_MENU = auto()
    MODEL_MENU = auto()
    MIN = auto()
    MAX = auto()
    CLOSE = auto()
    NORMAL = auto()


@lru_cache(maxsize=20)
def _icon_for_id(icon_id: TitleIconId, color: str) -> QtGui.QIcon:
    if icon_id is TitleIconId.WINDOW_MENU:
        return _get_icon("material-symbols:menu", color=color)
    elif icon_id is TitleIconId.MODEL_MENU:
        return _get_icon("octicon:ai-model-16", color=color)
    elif icon_id is TitleIconId.MIN:
        return _get_icon("material-symbols:minimize-rounded", color=color)
    elif icon_id is TitleIconId.MAX:
        return _get_icon("material-symbols:crop-5-4-outline", color=color)
    elif icon_id is TitleIconId.CLOSE:
        return _get_icon("material-symbols:close-rounded", color=color)
    elif icon_id is TitleIconId.NORMAL:
        return _get_icon(
            "material-symbols:filter-none-outline-rounded", rotate=180, color=color
        )
    else:
        raise ValueError(f"Invalid icon id: {icon_id}")


class QCentralWidget(QtW.QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        layout = QtW.QVBoxLayout(self)
        layout.setContentsMargins(1, 0, 1, 1)
        layout.setSpacing(0)


class QSubWindow(QtW.QMdiSubWindow):
    state_change_requested = QtCore.Signal(WindowState)
    rename_requested = QtCore.Signal(str)
    close_requested = QtCore.Signal()

    def __init__(self, widget: QtW.QWidget, title: str, icon_color: str = "black"):
        super().__init__()
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setAttribute(Qt.WidgetAttribute.WA_Hover, True)

        self._window_state = WindowState.NORMAL
        self._resize_state = ResizeState.NONE
        self._widget = widget
        self._current_icon_color = icon_color

        self._central_widget = QCentralWidget(self)
        self.setWidget(self._central_widget)

        self._title_bar = QSubWindowTitleBar(self, title)

        self._central_widget.layout().addWidget(self._title_bar)
        spacer = QtW.QWidget()
        layout = QtW.QVBoxLayout(spacer)
        layout.setContentsMargins(4, 4, 4, 4)
        layout.addWidget(widget)
        self._central_widget.layout().addWidget(spacer)
        self._last_geometry = self.geometry()
        self._anim_geometry = QtCore.QPropertyAnimation(self, b"geometry")
        self.setAcceptDrops(True)

        min_qsize = widget.minimumSizeHint().expandedTo(self._title_bar.minimumSize())
        max_qsize = widget.maximumSize()
        self._min_size = Size(min_qsize.width(), min_qsize.height())
        self._max_size = Size(max_qsize.width(), max_qsize.height())

    def _qt_mdiarea(self) -> QSubWindowArea:
        parent = self
        while parent is not None:
            parent = parent.parentWidget()
            if isinstance(parent, QSubWindowArea):
                return parent
        raise ValueError("Could not find the Qt main window.")

    def windowTitle(self) -> str:
        return self._title_bar._title_label.text()

    def setWindowTitle(self, title: str):
        self._title_bar._title_label.setText(title)

    def _my_wrapper(self) -> SubWindow:
        return self._widget._himena_widget

    @property
    def state(self) -> WindowState:
        return self._window_state

    def _set_icon_color(self, color: str):
        self._title_bar._set_icon_color(color)
        self._current_icon_color = color

    def _update_window_state(self, state: WindowState, animate: bool = True):
        state = WindowState(state)
        g_last = Size(self._last_geometry.width(), self._last_geometry.height())
        if self._window_state == state:
            return None
        if self._qt_mdiarea().viewMode() != QtW.QMdiArea.ViewMode.SubWindowView:
            self._window_state = state
            return None
        if animate:
            _setter = self._set_geometry_animated
        else:
            _setter = self.setGeometry
        g_parent = self.parentWidget().geometry()
        g_size = Size(g_parent.width(), g_parent.height())
        if state == WindowState.MIN:
            if self._window_state is WindowState.NORMAL:
                self._store_current_geometry()
            self.resize(124, self._title_bar.height() + 8)
            n_minimized = sum(
                1
                for sub_window in self._qt_mdiarea().subWindowList()
                if sub_window._window_state is WindowState.MIN
            )
            self._set_minimized(g_parent, n_minimized)
        elif state == WindowState.MAX:
            if self._window_state is WindowState.MIN:
                size_old = g_last
            elif self._window_state is WindowState.NORMAL:
                size_old = Size(self.size().width(), self.size().height())
                self._store_current_geometry()
            else:
                size_old = g_size
            _setter(g_parent)
            self._title_bar._toggle_size_btn.setIcon(
                _icon_for_id(TitleIconId.NORMAL, self._current_icon_color)
            )
            self._widget.setVisible(True)
            _checker.call_widget_resized_callback(
                self._my_wrapper().widget, size_old, g_size
            )
        elif state == WindowState.NORMAL:
            if self._window_state is WindowState.MIN:
                size_old = g_last
            else:
                size_old = g_size
            _setter(self._last_geometry)
            self._title_bar._toggle_size_btn.setIcon(
                _icon_for_id(TitleIconId.MAX, self._current_icon_color)
            )
            self._widget.setVisible(True)
            self._title_bar._fix_position()
            _checker.call_widget_resized_callback(
                self._my_wrapper().widget, size_old, g_last
            )
        elif state == WindowState.FULL:
            if self._window_state is WindowState.MIN:
                size_old = g_last
            elif self._window_state is WindowState.NORMAL:
                size_old = Size(self.size().width(), self.size().height())
                self._store_current_geometry()
            else:
                size_old = g_size
            _setter(g_parent)
            _checker.call_widget_resized_callback(
                self._my_wrapper().widget, size_old, g_last
            )
        else:  # pragma: no cover
            raise RuntimeError(f"Invalid window state value: {state}")

        self._title_bar._window_menu_btn.setVisible(state is not WindowState.MIN)
        self._title_bar._model_menu_btn.setVisible(state is not WindowState.MIN)
        self._title_bar.setVisible(state is not WindowState.FULL)
        self._title_bar._minimize_btn.setVisible(state is not WindowState.MIN)
        self._widget.setVisible(state is not WindowState.MIN)
        self._window_state = state
        return None

    def _store_current_geometry(self):
        self._last_geometry = self.geometry()
        _LOGGER.debug("Storing current geometry %r", self._last_geometry)

    def _set_minimized(self, geometry: QtCore.QRect, number: int = 0):
        self.move(2, geometry.height() - (self._title_bar.height() + 8) * (number + 1))
        self._title_bar._toggle_size_btn.setIcon(
            _icon_for_id(TitleIconId.NORMAL, self._current_icon_color)
        )
        self._widget.setVisible(False)

    def is_current(self) -> bool:
        return self._title_bar.property("isCurrent")

    def set_is_current(self, is_current: bool):
        """Set the isCurrent state of the sub-window and update styles."""
        self._title_bar.setProperty("isCurrent", is_current)
        self._title_bar.style().unpolish(self._title_bar)
        self._title_bar.style().polish(self._title_bar)

    def event(self, a0: QtCore.QEvent) -> bool:
        if a0.type() == QtCore.QEvent.Type.HoverMove:
            self._mouse_hover_event(a0.pos())
        elif a0.type() == QtCore.QEvent.Type.MouseButtonPress:
            assert isinstance(a0, QtGui.QMouseEvent)
            self._resize_state = self._check_resize_state(a0.pos())
        elif a0.type() == QtCore.QEvent.Type.MouseButtonRelease:
            self._resize_state = ResizeState.NONE
        return super().event(a0)

    def _check_resize_state(
        self, mouse_pos: QtCore.QPoint, thickness: int = 6, thickness_top: int = 4
    ) -> ResizeState:
        is_left = mouse_pos.x() < thickness
        is_right = mouse_pos.x() > self.width() - thickness
        is_top = mouse_pos.y() < thickness_top
        is_bottom = mouse_pos.y() > self.height() - thickness
        return ResizeState.from_bools(is_left, is_right, is_top, is_bottom)

    @qthrottled(timeout=10)
    def _mouse_hover_event(self, event_pos: QtCore.QPoint):
        # if the cursor is at the edges, set the cursor to resize
        if self._window_state is not WindowState.NORMAL:
            return None
        resize_state = self._check_resize_state(event_pos)
        current_button = QtW.QApplication.mouseButtons()
        if current_button == Qt.MouseButton.NoButton:
            self.setCursor(CURSOR_SHAPE_MAP[resize_state])
        elif current_button & Qt.MouseButton.LeftButton:
            # NOTE: Method "minimusSizeHint" represents the minimum size of the widget
            # as a window
            rect = self.geometry()
            if new_rect := self._resize_state.resize_widget(
                WindowRect(rect.left(), rect.top(), rect.width(), rect.height()),
                mouse_pos=(event_pos.x(), event_pos.y()),
                min_size=self._min_size,
                max_size=self._max_size,
            ):
                subwin = self._my_wrapper()
                size_old = Size(rect.width(), rect.height())

                # update window rect
                subwin.rect = new_rect

                # update window anchor
                main_qsize = self._qt_mdiarea().size()
                subwin.anchor = subwin.anchor.update_for_window_rect(
                    (main_qsize.width(), main_qsize.height()),
                    new_rect,
                )
                _checker.call_widget_resized_callback(
                    self._my_wrapper().widget,
                    size_old,
                    new_rect.size(),
                )
        return None

    def _set_geometry_animated(self, rect: QtCore.QRect):
        if self._anim_geometry.state() == QtCore.QAbstractAnimation.State.Running:
            self._anim_geometry.stop()
        self._anim_geometry.setTargetObject(self)
        self._anim_geometry.setPropertyName(b"geometry")
        self._anim_geometry.setStartValue(QtCore.QRect(self.geometry()))
        self._anim_geometry.setEndValue(rect)
        self._anim_geometry.setDuration(60)
        self._anim_geometry.start()

    def _pixmap_resized(
        self,
        size: QtCore.QSize,
        outline: QtGui.QColor | None = None,
    ) -> QtGui.QPixmap:
        return pixmap_resized(self, size, outline)

    def _find_me(self) -> tuple[int, int]:
        return self._find_me_and_main()[0]

    def _find_me_and_main(self) -> tuple[tuple[int, int], MainWindowQt]:
        main = get_main_window(self)
        for i_tab, tab in main.tabs.enumerate():
            for i_win, win in tab.enumerate():
                _, qwidget = win._split_interface_and_frontend()
                if qwidget is self._widget:
                    return (i_tab, i_win), main
        raise RuntimeError("Could not find the sub-window in the main window.")

    def dragEnterEvent(self, a0: QtGui.QDragEnterEvent | None) -> None:
        if a0 is None:
            return None
        if model := _drag.get_dragging_model():
            if self._my_wrapper()._is_drop_accepted(model):
                a0.accept()
                return None
        a0.ignore()
        return None

    def dropEvent(self, a0: QtGui.QDropEvent | None) -> None:
        if a0 is None:
            return None
        if model := _drag.drop():
            # user defined drop event
            self_wrapper = self._my_wrapper()
            if isinstance(src := a0.source(), QSubWindow):
                source = src._my_wrapper()
            else:
                source = None
            if self_wrapper.widget is src:  # same widget
                a0.ignore()
                _LOGGER.info("dropping to the same widget, ignore it")
                return None

            if self_wrapper._process_drop_event(model, source):
                a0.accept()
                return None
        a0.ignore()
        a0.setDropAction(Qt.DropAction.IgnoreAction)
        return None


class QTitleBarToolButton(QtW.QToolButton):
    """Tool button for the title bar of the sub-window."""

    def __init__(
        self,
        icon_id: TitleIconId,
        color: str,
        tooltip: str,
        callback: Callable[[], None],
    ):
        super().__init__()
        self._icon_id = icon_id
        self._current_icon_color = color
        self.setToolTip(tooltip)
        self.clicked.connect(callback)
        self._update_icon_color(color)

    def _set_size(self, size: int):
        self.setFixedSize(size, size)
        self.setIconSize(QtCore.QSize(size - 1, size - 1))

    def _update_icon_color(self, color: str):
        self._current_icon_color = color
        self.setIcon(_icon_for_id(self._icon_id, color))


class QDummyToolButton(QtW.QWidget):
    """Invisible widget just for API compatibility."""

    def _set_size(self, size: int):
        self.setFixedSize(size, size)

    def _update_icon_color(self, color: str):
        pass


class QSubWindowTitleBar(QtW.QFrame):
    """The subwindow title bar widget."""

    def __init__(self, subwindow: QSubWindow, title: str):
        super().__init__()
        self._drag_position: QtCore.QPoint | None = None
        self._is_window_drag_mode: bool = False
        self._is_double_clicking: bool = False
        self._resize_position: QtCore.QPoint | None = None
        self._subwindow = subwindow

        self.setFrameShape(QtW.QFrame.Shape.StyledPanel)
        self.setFrameShadow(QtW.QFrame.Shadow.Raised)
        self.setMinimumWidth(100)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

        self._window_menu_btn = QTitleBarToolButton(
            icon_id=TitleIconId.WINDOW_MENU,
            color="black",
            tooltip="Menu for this window",
            callback=self._show_window_menu,
        )
        if self._get_model_type() is not None:
            self._model_menu_btn = QTitleBarToolButton(
                icon_id=TitleIconId.MODEL_MENU,
                color="black",
                tooltip="Menu specific to the model",
                callback=self._show_model_menu,
            )
        else:
            self._model_menu_btn = QDummyToolButton()
            self._model_menu_btn.setVisible(False)

        self._index_label = QtW.QLabel()
        self._index_label.setObjectName("indexLabel")
        self._index_label.setAlignment(
            Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter
        )
        _index_font = self._index_label.font()
        _index_font.setPointSize(8)
        _index_font.setBold(True)
        self._index_label.setFont(_index_font)
        self._index_label.setFixedWidth(20)
        self._index_label.setContentsMargins(0, 0, 0, 0)
        self._index_label.setSizePolicy(
            QtW.QSizePolicy.Policy.Fixed, QtW.QSizePolicy.Policy.Fixed
        )

        self._title_label = QtW.QLabel(title)
        self._title_label.setIndent(3)
        self._title_label.setContentsMargins(0, 0, 0, 0)
        self._title_label.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Fixed
        )

        self._line_edit = QRenameLineEdit(self._title_label)

        @self._line_edit.rename_requested.connect
        def _(new_name: str):
            self._subwindow.rename_requested.emit(new_name)

        self._minimize_btn = QTitleBarToolButton(
            icon_id=TitleIconId.MIN,
            color="black",
            tooltip="Minimize this window",
            callback=self._minimize,
        )
        self._toggle_size_btn = QTitleBarToolButton(
            icon_id=TitleIconId.MAX,
            color="black",
            tooltip="Toggle the size of this window",
            callback=self._toggle_size,
        )
        self._close_btn = QTitleBarToolButton(
            icon_id=TitleIconId.CLOSE,
            color="black",
            tooltip="Close this window",
            callback=self._close,
        )

        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(1)
        layout.addWidget(self._window_menu_btn)
        layout.addWidget(self._model_menu_btn)
        layout.addWidget(self._index_label)
        layout.addWidget(self._title_label)
        layout.addWidget(self._minimize_btn, alignment=Qt.AlignmentFlag.AlignRight)
        layout.addWidget(self._toggle_size_btn, alignment=Qt.AlignmentFlag.AlignRight)
        layout.addWidget(self._close_btn, alignment=Qt.AlignmentFlag.AlignRight)

        self.setProperty("isCurrent", False)
        self.setAcceptDrops(True)
        self._bar_size = 18
        self._set_bar_size(self._bar_size)

    def _get_model_type(self) -> str | None:
        interf = self._subwindow._my_wrapper().widget
        if hasattr(interf, "model_type"):
            try:
                return interf.model_type()
            except Exception:
                return None
        elif hasattr(interf, "__himena_model_type__"):
            return interf.__himena_model_type__
        return None

    def _set_visible(self, visible: bool):
        if visible:
            self._set_bar_size(self._bar_size)
        else:
            self._set_bar_size(0)

    def _set_bar_size(self, height: int):
        self.setFixedHeight(height)
        self._index_label.setFixedHeight(height)
        self._title_label.setFixedHeight(height)
        if height > 1:
            self._window_menu_btn._set_size(height - 1)
            self._model_menu_btn._set_size(height - 1)
            self._minimize_btn._set_size(height - 1)
            self._toggle_size_btn._set_size(height - 1)
            self._close_btn._set_size(height - 1)

    def _set_icon_color(self, color):
        for toolbtn in (
            self._window_menu_btn,
            self._model_menu_btn,
            self._minimize_btn,
            self._toggle_size_btn,
            self._close_btn,
        ):
            toolbtn._update_icon_color(color)

    def _start_renaming(self):
        self._line_edit.show()
        self._move_line_edit(self._title_label.rect(), self._title_label.text())

    def _move_line_edit(
        self,
        rect: QtCore.QRect,
        text: str,
    ) -> QtW.QLineEdit:
        geometry = self._line_edit.geometry()
        geometry.setWidth(rect.width())
        geometry.setHeight(rect.height())
        geometry.moveCenter(rect.center())
        self._line_edit.setGeometry(geometry)
        self._line_edit.setText(text)
        self._line_edit.setHidden(False)
        self._line_edit.setFocus()
        self._line_edit.selectAll()

    def _minimize(self):
        self._subwindow.state_change_requested.emit(WindowState.MIN)

    def _toggle_size(self):
        if self._subwindow._window_state is WindowState.NORMAL:
            self._subwindow.state_change_requested.emit(WindowState.MAX)
        else:
            self._subwindow.state_change_requested.emit(WindowState.NORMAL)

    def _maximize(self):
        self._subwindow.state_change_requested.emit(WindowState.MAX)

    def _toggle_full_screen(self):
        if self._subwindow._window_state is WindowState.FULL:
            self._subwindow.state_change_requested.emit(WindowState.NORMAL)
        else:
            self._subwindow.state_change_requested.emit(WindowState.FULL)

    def _close(self):
        return self._subwindow.close_requested.emit()

    def _make_tooltip(self):
        """Make the tooltip for the title bar"""
        qwin = self._subwindow
        attrs: list[str] = [f"<b>Title</b>: {self._title_label.text()}"]
        if _model_type := self._get_model_type():
            attrs.append(f"<b>Type</b>: {_model_type}")
        attrs.append(
            f"<b>Widget</b>: {get_display_name(qwin._widget.__class__, sep=' ')}"
        )
        sub = qwin._my_wrapper()
        attrs.append(f"<b>Save behavior</b>: {sub.save_behavior!r}")
        tooltip = "<br>".join(attrs)
        return tooltip

    def event(self, a0: QtCore.QEvent) -> bool:
        if a0.type() == QtCore.QEvent.Type.ToolTip:
            a0 = cast(QtGui.QHelpEvent, a0)
            QtW.QToolTip.showText(a0.globalPos(), self._make_tooltip(), self)
            return True
        return super().event(a0)

    # drag events for moving the window
    def mousePressEvent(self, event: QtGui.QMouseEvent):
        self._is_window_drag_mode = False
        if self._is_double_clicking:
            return super().mousePressEvent(event)
        _subwin = self._subwindow
        _left_btn = event.buttons() & Qt.MouseButton.LeftButton
        _middle_btn = event.buttons() & Qt.MouseButton.MiddleButton
        _ctrl_mod = event.modifiers() & Qt.KeyboardModifier.ControlModifier
        if (_left_btn and _ctrl_mod) or _middle_btn:
            # start dragging subwindow
            self._is_window_drag_mode = True
            drag = self._make_subwindow_drag()
            self._subwindow.hide()
            try:
                drag.exec()
            finally:
                # NOTE: if subwindow is dropped to another tab, the old one will be
                # removed from the MdiArea. In this case, the subwindow should not
                # be shown again.
                if self._subwindow.parent():
                    self._subwindow.show()
        else:
            if _subwin._window_state == WindowState.MIN:
                # cannot move minimized window
                return
        self._store_drag_position(event.globalPos())
        return super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        _subwin = self._subwindow
        if event.buttons() == Qt.MouseButton.NoButton:
            self._drag_position = None
        elif (
            event.buttons() == Qt.MouseButton.LeftButton
            and self._drag_position is not None
            and _subwin._resize_state is ResizeState.NONE
            and not self._is_window_drag_mode
        ):
            g = _subwin.geometry()
            main_qsize = _subwin._qt_mdiarea().size()
            wrapper = _subwin._my_wrapper()
            if _subwin._window_state == WindowState.MAX:
                # change to normal without moving
                self._toggle_size_btn.setIcon(
                    _icon_for_id(TitleIconId.MAX, self._subwindow._current_icon_color)
                )
                _subwin._widget.setVisible(True)
                _subwin._window_state = WindowState.NORMAL
                # restore the last size
                _last_geo = _subwin._last_geometry
                _next_rect = WindowRect(
                    event.pos().x() - _last_geo.width() // 2,
                    event.pos().y() - self.height() // 2,
                    _last_geo.width(),
                    _last_geo.height(),
                )
                self._store_drag_position(event.globalPos())
            else:
                # drag the subwindow
                new_pos = event.globalPos() - self._drag_position
                offset = self.height() - 4
                if new_pos.y() < -offset:
                    new_pos.setY(-offset)
                _next_rect = wrapper.rect.move_top_left(new_pos.x(), new_pos.y())

            # update window anchor
            wrapper.rect = _next_rect
            wrapper.anchor = wrapper.anchor.update_for_window_rect(
                (main_qsize.width(), main_qsize.height()),
                WindowRect.from_tuple(g.left(), g.top(), g.width(), g.height()),
            )
        return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        self._is_window_drag_mode = False
        self._is_double_clicking = False
        if event.button() == Qt.MouseButton.LeftButton:
            self._drag_position = None
            self._fix_position()
        elif event.button() == Qt.MouseButton.RightButton:
            # context menu
            if (event.pos() - self._drag_position).manhattanLength() < 10:
                context_menu = self._prep_window_menu()
                context_menu.exec(event.globalPos())
        return super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent):
        self._is_double_clicking = True
        if event.buttons() & Qt.MouseButton.LeftButton:
            self._toggle_size()
        return super().mouseDoubleClickEvent(event)

    def wheelEvent(self, event: QtGui.QWheelEvent):
        if event.modifiers() == Qt.KeyboardModifier.NoModifier:
            self._wheel_event(event.angleDelta().y())
        return super().wheelEvent(event)

    def _wheel_event(self, dy: int):
        subwin = self._subwindow
        i_tab, i_win = subwin._find_me()
        main = get_main_window(subwin)
        sub = main.tabs[i_tab][i_win]
        size_old = sub.size
        if dy > 0:
            rect_new = sub.rect.resize_relative(1.1, 1.1)
        else:
            rect_new = sub.rect.resize_relative(1 / 1.1, 1 / 1.1)
        inst = main._instructions.updated(animate=False)
        sub._set_rect(rect_new, inst)
        size_new = rect_new.size()
        return _checker.call_widget_resized_callback(
            subwin._my_wrapper().widget, size_old, size_new
        )

    def _make_subwindow_drag(self):
        _subwin = self._subwindow
        drag = QtGui.QDrag(_subwin)
        drag.destroyed.connect(_drag.clear)
        mime_data = QtCore.QMimeData()
        _wrapper = self._subwindow._my_wrapper()
        if isinstance(
            _meth := _wrapper._widget_workflow.last(),
            LocalReaderMethod,
        ):
            if isinstance(_meth.path, Path):
                mime_data.setUrls([QtCore.QUrl(_meth.path.as_uri())])
            elif isinstance(_meth.path, list):
                mime_data.setUrls([QtCore.QUrl(fp.as_uri()) for fp in _meth.path])
        drag.setMimeData(mime_data)
        if _wrapper.supports_to_model:
            model = DragDataModel(
                getter=_wrapper.to_model,
                type=_wrapper.model_type(),
            )
            _drag.drag(model)
        drag.setPixmap(_subwin._pixmap_resized(QtCore.QSize(150, 150)))
        return drag

    def dragEnterEvent(self, a0: QtGui.QDragEnterEvent | None) -> None:
        return self._subwindow.dragEnterEvent(a0)

    def dropEvent(self, a0: QtGui.QDropEvent | None) -> None:
        return self._subwindow.dropEvent(a0)

    def enterEvent(self, event):
        ui = get_main_window(self)
        ui.set_status_tip("`Ctrl+LeftButton` or `MiddleButton` to drag the window")
        return super().enterEvent(event)

    def leaveEvent(self, a0):
        try:
            main = get_main_window(self)
        except ValueError:
            pass
        else:
            main.set_status_tip("")
        return super().leaveEvent(a0)

    def _store_drag_position(self, global_pos: QtCore.QPoint):
        subwin = self._subwindow
        self._drag_position = global_pos - subwin.frameGeometry().topLeft()

    def _fix_position(self):
        self_pos = self.mapToGlobal(self._subwindow.rect().topRight())
        parent_pos = self._subwindow.parentWidget().mapToGlobal(QtCore.QPoint(0, 0))
        if self_pos.y() < parent_pos.y():
            self._subwindow.move(self._subwindow.pos().x(), 0)
        if self_pos.x() < parent_pos.x():
            self._subwindow.move(0, self._subwindow.pos().y())

    def _show_model_menu(self):
        context_menu = self._prep_model_menu()
        return self._exec_menu_at_button(context_menu, self._model_menu_btn)

    def _prep_window_menu(self) -> QtW.QMenu:
        main = get_main_window(self)
        app = main._model_app

        context_menu = build_qmodel_menu(MenuId.WINDOW, app=app.name, parent=self)
        ctx = main._ctx_keys
        ctx._update(main)
        context_menu.update_from_context(ctx.dict())
        return context_menu

    def _prep_model_menu(self) -> QtW.QMenu:
        model_type = self._get_model_type()
        if model_type is None:
            return None
        ui = get_main_window(self)
        model_subtypes = model_type.split(".")
        supertype_menus: list[QtW.QMenu] = []
        _id = f"/model_menu:{model_type}"

        context_menu = build_qmodel_menu(_id, app=ui.model_app.name, parent=self)
        ctx = ui._ctx_keys
        ctx._update(ui)
        ctx_dict = ctx.dict()
        open_in_actions: list[QCommandRuleAction] = []
        _n_enabled = 0
        for i in range(1, len(model_subtypes) + 1):
            _typ = ".".join(model_subtypes[:i])
            _id_open_in = f"/open-in/{_typ}"
            for menu_group in ui.model_app.menus.iter_menu_groups(_id_open_in):
                for menu_or_submenu in menu_group:
                    if isinstance(menu_or_submenu, MenuItem):
                        action = QCommandRuleAction(
                            menu_or_submenu.command, ui.model_app.name, self
                        )
                        is_enabled = menu_or_submenu.command.enablement.eval(ctx_dict)
                        _n_enabled += int(is_enabled)
                        action.setEnabled(is_enabled)
                        open_in_actions.append(action)
        if open_in_actions and _n_enabled > 0:
            open_in_menu = QtW.QMenu()
            open_in_menu.setParent(context_menu, open_in_menu.windowFlags())
            open_in_menu.setTitle("Open in ...")
            actions = context_menu.actions()
            if len(actions) == 0:
                context_menu.addMenu(open_in_menu)
            else:
                context_menu.insertMenu(actions[0], open_in_menu)
            for action in open_in_actions:
                open_in_menu.addAction(action)

        # also add supertype actions
        for num in range(1, len(model_subtypes)):
            _typ = ".".join(model_subtypes[:num])
            _id = f"/model_menu:{_typ}"
            _menu = build_qmodel_menu(_id, app=ui.model_app.name, parent=self)
            _menu.setTitle(f'"{_typ}" menu')
            supertype_menus.append(_menu)
        if supertype_menus:
            context_menu.addSeparator()
            for menu in supertype_menus:
                if not menu.isEmpty():
                    context_menu.addMenu(menu)
        context_menu.update_from_context(ctx_dict)
        return context_menu

    def _show_window_menu(self):
        context_menu = self._prep_window_menu()
        return self._exec_menu_at_button(context_menu, self._window_menu_btn)

    def _exec_menu_at_button(self, menu: QtW.QMenu, btn: QtW.QToolButton):
        pos_local = btn.rect().bottomLeft()
        pos_global = btn.mapToGlobal(pos_local)
        menu.exec(pos_global)
        return None


def pixmap_resized(
    widget: QtW.QWidget,
    size: QtCore.QSize,
    outline: QtGui.QColor | None = None,
) -> QtGui.QPixmap:
    pixmap = widget.grab().scaled(
        size,
        Qt.AspectRatioMode.KeepAspectRatio,
        Qt.TransformationMode.SmoothTransformation,
    )
    if outline is not None:
        painter = QtGui.QPainter(pixmap)
        painter.setPen(QtGui.QPen(outline, 2))
        painter.drawRect(pixmap.rect().adjusted(0, 0, -1, -1))
        painter.end()
    return pixmap


CURSOR_SHAPE_MAP = {
    ResizeState.TOP: Qt.CursorShape.SizeVerCursor,
    ResizeState.BOTTOM: Qt.CursorShape.SizeVerCursor,
    ResizeState.LEFT: Qt.CursorShape.SizeHorCursor,
    ResizeState.RIGHT: Qt.CursorShape.SizeHorCursor,
    ResizeState.TOP_LEFT: Qt.CursorShape.SizeFDiagCursor,
    ResizeState.TOP_RIGHT: Qt.CursorShape.SizeBDiagCursor,
    ResizeState.BOTTOM_LEFT: Qt.CursorShape.SizeBDiagCursor,
    ResizeState.BOTTOM_RIGHT: Qt.CursorShape.SizeFDiagCursor,
    ResizeState.NONE: Qt.CursorShape.ArrowCursor,
}


def get_subwindow(widget: QtW.QWidget) -> QSubWindow:
    window = widget.parentWidget().parentWidget().parentWidget()
    if not isinstance(window, QSubWindow):
        raise ValueError(f"Widget {widget!r} is not in a sub-window.")
    return window



from __future__ import annotations

from typing import TYPE_CHECKING
import numpy as np
from qtpy import QtWidgets as QtW
from qtpy import QtCore, QtGui
from superqt import QLabeledDoubleSlider, QToggleSwitch, QElidingLabel
from superqt.utils import qthrottled

from himena.qt._utils import qsignal_blocker
from himena_builtins.qt.widgets._image_components import QHistogramView
from himena.utils.enum import StrEnum

if TYPE_CHECKING:
    from himena_builtins.qt.widgets.image import QImageViewBase, ChannelInfo


class ImageType(StrEnum):
    SINGLE = "Single"
    RGB = "RGB"
    MULTI = "Multi"
    OTHERS = "Others"


class ComplexMode(StrEnum):
    REAL = "Real"
    IMAG = "Imag"
    ABS = "Abs"
    LOG_ABS = "Log Abs"
    PHASE = "Phase"


class ChannelMode(StrEnum):
    COMP = "Comp."
    MONO = "Mono"
    GRAY = "Gray"


class RGBMode(StrEnum):
    COLOR = "Color"
    GRAY = "Gray"


class QImageViewControlBase(QtW.QWidget):
    def __init__(self, image_view: QImageViewBase):
        super().__init__()
        self._image_view = image_view
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self._interp_check_box = QToggleSwitch()
        self._interp_check_box.setText("smooth")
        self._interp_check_box.setChecked(False)
        self._interp_check_box.setMaximumHeight(36)
        self._interp_check_box.toggled.connect(self._interpolation_changed)

        self._hover_info = QElidingLabel()
        self._hover_info.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter
        )
        self._hover_info.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )

        for wdt in self._widgets_to_add():
            layout.addWidget(wdt)

    def _widgets_to_add(self) -> list[QtW.QWidget]:
        return [self._hover_info, self._interp_check_box]

    def _interpolation_changed(self, checked: bool):
        self._image_view._img_view.setSmoothing(checked)

    def update_rgb_channel_dtype(self, is_rgb: bool, nchannels: int, dtype):
        pass


class QImageViewControl(QImageViewControlBase):
    def __init__(self, image_view: QImageViewBase):
        self._complex_mode_old = "Abs"
        self._cmp_mode_combo = QtW.QComboBox()
        self._cmp_mode_combo.addItems(
            [ComplexMode.REAL, ComplexMode.IMAG, ComplexMode.ABS, ComplexMode.LOG_ABS,
             ComplexMode.PHASE]
        )  # fmt: skip
        self._cmp_mode_combo.setCurrentIndex(2)
        self._cmp_mode_combo.setToolTip("Method to display complex data")

        self._chn_vis = QChannelToggleSwitches()

        self._chn_mode_combo = QtW.QComboBox()
        self._chn_mode_combo.addItems([""])
        self._chn_mode_combo.setToolTip("Method to display multi-channel data")
        self._image_type = ImageType.OTHERS

        self._auto_cont_btn = QtW.QPushButton("Auto")
        self._auto_cont_btn.setToolTip("Auto contrast")

        self._histogram = QHistogramView()
        self._histogram.setFixedWidth(120)
        super().__init__(image_view)
        self._cmp_mode_combo.hide()
        self._chn_mode_combo.hide()
        self._cmp_mode_combo.currentTextChanged.connect(self._on_complex_mode_change)
        self._chn_vis.stateChanged.connect(self._on_channel_visibility_change)
        self._histogram.clim_changed.connect(self._clim_changed)
        self._auto_cont_btn.clicked.connect(self._auto_contrast)
        self._chn_mode_combo.currentTextChanged.connect(self._on_channel_mode_change)

    def _widgets_to_add(self) -> list[QtW.QWidget]:
        return [
            self._hover_info, self._cmp_mode_combo, self._chn_vis,
            self._chn_mode_combo, self._auto_cont_btn, self._histogram,
            self._interp_check_box,
        ]  # fmt: skip

    def update_rgb_channel_dtype(self, is_rgb: bool, nchannels: int, dtype):
        dtype = np.dtype(dtype)
        if is_rgb:
            kind = ImageType.RGB
        elif nchannels > 1:
            kind = ImageType.MULTI
        else:
            kind = ImageType.SINGLE
        if kind != self._image_type:
            if kind is ImageType.RGB:
                self._chn_mode_combo.clear()
                self._chn_mode_combo.addItems([RGBMode.COLOR, RGBMode.GRAY])
                self._chn_mode_combo.show()
                self._chn_vis.hide()
            elif kind is ImageType.MULTI:
                self._chn_mode_combo.clear()
                self._chn_mode_combo.addItems(
                    [ChannelMode.COMP, ChannelMode.MONO, ChannelMode.GRAY]
                )
                self._chn_mode_combo.show()
                self._chn_vis.show()
            else:
                self._chn_mode_combo.clear()
                self._chn_mode_combo.addItems([""])
                self._chn_mode_combo.hide()
                self._chn_vis.hide()
            self._image_type = kind
            self._chn_mode_combo.setCurrentIndex(0)
        self._cmp_mode_combo.setVisible(dtype.kind == "c")
        if dtype.kind in "uib":
            self._histogram.setValueFormat(".0f")
        else:
            self._histogram.setValueFormat(".3g")
        return None

    def complex_transform(self, arr: np.ndarray) -> np.ndarray:
        """Transform complex array according to the current complex mode."""
        if self._cmp_mode_combo.currentText() == ComplexMode.REAL:
            return arr.real
        if self._cmp_mode_combo.currentText() == ComplexMode.IMAG:
            return arr.imag
        if self._cmp_mode_combo.currentText() == ComplexMode.ABS:
            return np.abs(arr)
        if self._cmp_mode_combo.currentText() == ComplexMode.LOG_ABS:
            return np.log(np.abs(arr) + 1e-6)
        if self._cmp_mode_combo.currentText() == ComplexMode.PHASE:
            return np.angle(arr)
        return arr

    @qthrottled(timeout=100)
    def _clim_changed(self, clim: tuple[float, float]):
        view = self._image_view
        ch = view.current_channel()
        ch.clim = clim
        idx = ch.channel_index or 0
        imtup = view._current_image_slices[idx]
        with qsignal_blocker(self._histogram):
            _grays = (RGBMode.GRAY, ChannelMode.GRAY)
            if imtup.visible:
                arr = ch.transform_image(
                    view._current_image_slices[idx].arr,
                    complex_transform=self.complex_transform,
                    is_rgb=view._is_rgb,
                    is_gray=self._chn_mode_combo.currentText() in _grays,
                )
            else:
                arr = None
            view._img_view.set_array(idx, arr)

    def _on_channel_mode_change(self, mode: str):
        self._chn_vis.setVisible(mode == ChannelMode.COMP)
        self._on_channel_visibility_change()

    def _channel_visibility(self) -> list[bool]:
        caxis = self._image_view._channel_axis
        if caxis is None:
            return [True]  # No channels, always visible
        is_composite = self._chn_mode_combo.currentText() == ChannelMode.COMP
        if is_composite:
            visibilities = self._chn_vis._check_states()
        else:
            visibilities = [False] * len(self._chn_vis._toggle_switches)
            sl = self._image_view._dims_slider.value()
            ith_channel = sl[caxis]
            if len(visibilities) <= ith_channel:
                return [True] * len(sl)  # before initialization
            visibilities[ith_channel] = True
        return visibilities

    def _on_channel_visibility_change(self):
        visibilities = self._channel_visibility()
        self._image_view._update_image_visibility(visibilities)

    def _on_complex_mode_change(self):
        cur = self._cmp_mode_combo.currentText()
        self._image_view._reset_image()
        self._complex_mode_old = cur
        # TODO: auto contrast and update colormap

    def _auto_contrast(self):
        min_ = float("inf")
        max_ = -float("inf")
        for item in self._histogram._hist_items:
            min_ = min(item._edges[0], min_)
            max_ = max(item._edges[-1], max_)
        if np.isinf(min_) or np.isinf(max_):
            return

        view = self._image_view
        sl = view._dims_slider.value()
        img_slice = view._get_image_slice_for_channel(sl)
        if img_slice.dtype.kind == "c":
            img_slice = self.complex_transform(img_slice)
        ch = view.current_channel(sl)
        ch.clim = (min_, max_)
        self._histogram.set_view_range(min_, max_)
        view._set_image_slice(img_slice, ch)
        self._histogram.set_clim((min_, max_))


class QImageLabelViewControl(QImageViewControlBase):
    def __init__(self, image_view: QImageViewBase):
        self._opacity_slider = QLabeledDoubleSlider(QtCore.Qt.Orientation.Horizontal)
        self._opacity_slider.setValue(0.6)
        self._opacity_slider.setFixedWidth(120)
        self._opacity_slider.setRange(0.0, 1.0)
        self._opacity_slider.setToolTip("Opacity of the label colors")
        super().__init__(image_view)
        self._opacity_slider.valueChanged.connect(image_view._reset_image)

    def _widgets_to_add(self):
        return [self._hover_info, self._opacity_slider, self._interp_check_box]


class QChannelToggleSwitch(QToggleSwitch):
    """Toggle switch for channel visibility."""

    def __init__(self, channel: ChannelInfo):
        super().__init__()
        self._channel_info = channel
        self.setChecked(True)

    def set_channel(self, channel: ChannelInfo):
        self._channel_info = channel
        self.setText(channel.name)
        self._channel_on_color = QtGui.QColor.fromRgbF(*channel.colormap(0.5))

    def drawGroove(self, painter, rect, option):
        painter.setPen(QtCore.Qt.PenStyle.NoPen)
        is_checked = option.state & QtW.QStyle.StateFlag.State_On
        painter.setBrush(self._channel_on_color if is_checked else option.off_color)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, True)
        painter.setOpacity(0.65)
        painter.drawRect(rect)

    def drawHandle(self, painter, rect, option):
        painter.setPen(QtCore.Qt.PenStyle.SolidLine)
        is_checked = option.state & QtW.QStyle.StateFlag.State_On
        painter.setBrush(self._channel_on_color if is_checked else option.off_color)
        painter.setOpacity(1.0)
        painter.drawRect(rect)


class QChannelToggleSwitches(QtW.QScrollArea):
    stateChanged = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        central = QtW.QWidget()
        layout = QtW.QGridLayout(central)
        layout.setContentsMargins(2, 2, 2, 2)
        self._layout = layout
        self._toggle_switches: list[QChannelToggleSwitch] = []
        self.setWidget(central)
        self._label_font = QtGui.QFont("Arial", 8)

    def set_channels(self, channels: list[ChannelInfo]):
        labels = [ch.name for ch in channels]
        for ith in range(len(self._toggle_switches), len(labels)):
            sw = QChannelToggleSwitch(channels[ith])
            sw.setChecked(True)
            sw.setFont(self._label_font)
            sw.toggled.connect(self._emit_state_changed)
            row, col = divmod(ith, 3)
            self._layout.addWidget(sw, row, col)
            self._toggle_switches.append(sw)
        while len(self._toggle_switches) > len(labels):
            sw = self._toggle_switches.pop()
            sw.setParent(None)
        for i in range(len(channels)):
            sw = self._toggle_switches[i]
            sw.set_channel(channels[i])
        self.setFixedWidth(70 * min(3, len(self._toggle_switches)) + 10)

    def _emit_state_changed(self):
        self.stateChanged.emit()

    def _check_states(self) -> list[bool]:
        return [sw.isChecked() for sw in self._toggle_switches]



"""Builtin QtConsole plugin."""

from dataclasses import dataclass
from typing import Literal
from himena.plugins import (
    register_dock_widget_action,
    add_default_status_tip,
    config_field,
)

add_default_status_tip(
    short="command palette",
    long="Ctrl+Shift+P to search for all the registered commands",
)


@dataclass
class ConsoleConfig:
    """Configuration for the console."""

    main_window_symbol: str = config_field(
        default="ui",
        tooltip="Variable name used for the main window instance",
    )
    exit_app_from_console: bool = config_field(
        default=True,
        tooltip="Use the `exit` IPython magic to exit the application.",
    )
    matplotlib_backend: Literal["inline", "himena_builtins"] = config_field(
        default="inline",
        tooltip="Plot backend when the script is executed in the console.",
    )

    @property
    def mpl_backend(self) -> str:
        if self.matplotlib_backend == "himena_builtins":
            from himena_builtins.qt.plot import BACKEND_HIMENA

            return BACKEND_HIMENA
        return self.matplotlib_backend


@register_dock_widget_action(
    title="Console",
    area="bottom",
    keybindings=["Ctrl+Shift+C"],
    singleton=True,
    command_id="builtins:console",
    plugin_configs=ConsoleConfig(),
)
def install_console(ui):
    """Python interpreter widget."""
    from himena_builtins.qt.console._widget import QtConsole

    return QtConsole.get_or_create(ui)



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtCore, QtGui
from functools import singledispatch
from datetime import timedelta
from magicgui.widgets import request_values
from himena.consts import StandardType, MonospaceFontFamily
from himena.plugins import validate_protocol
from himena import workflow as _wf
from himena.qt._utils import drag_model, get_main_window
from himena_builtins.qt.widgets._table_components._selection_model import (
    SelectionModel,
    Index,
)
from himena.types import WidgetDataModel


class QWorkflowView(QtW.QWidget):
    """A widget that displays a workflow of another widget.

    A workflow is a directed acyclic graph of operations. Each operation is represented
    by a item in this widget. The item can be expanded to show the details. The parents
    of the selected item are highlighted with dashed lines.
    """

    def __init__(self):
        super().__init__()
        layout = QtW.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self._tree_widget = QWorkflowTree(_wf.Workflow())
        self._tree_widget.setColumnCount(1)
        self._tree_widget.setHeaderHidden(True)
        layout.addWidget(self._tree_widget)
        self._tree_widget.setFont(QtGui.QFont(MonospaceFontFamily))
        self._tree_widget.right_clicked.connect(self._on_right_clicked)
        self._modified = False
        self._editable = True

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        wf = model.value
        if not isinstance(wf, _wf.Workflow):
            raise ValueError(f"Expected Workflow, got {type(wf)}")
        self._tree_widget.set_workflow(wf)

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        return WidgetDataModel(
            value=self._tree_widget._workflow,
            type=self.model_type(),
            extension_default=".workflow.json",
        )

    @validate_protocol
    def model_type(self) -> str:
        if any(
            isinstance(step, _wf.UserInput)
            for step in self._tree_widget._workflow.steps
        ):
            return StandardType.WORKFLOW_PARAMETRIC
        return StandardType.WORKFLOW

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 400, 320

    @validate_protocol
    def is_modified(self) -> bool:
        return self._modified

    @validate_protocol
    def set_modified(self, modified: bool) -> None:
        self._modified = modified

    @validate_protocol
    def is_editable(self) -> bool:
        return self._editable

    @validate_protocol
    def set_editable(self, editable: bool) -> None:
        self._editable = editable

    def set_workflow(self, workflow: _wf.Workflow) -> None:
        """Set the workflow."""
        self._tree_widget.set_workflow(workflow)

    def _replace_with_file_reader(self, item: QtW.QTreeWidgetItem, how: str) -> None:
        row = self._tree_widget.indexOfTopLevelItem(item)
        step = self._tree_widget._workflow[row]
        wf_new = self._tree_widget._workflow.replace_with_input(step.id, how=how)
        self.set_workflow(wf_new)
        self._modified = True

    def _toggle_to_be_added(self, item: QtW.QTreeWidgetItem) -> None:
        row = self._tree_widget.indexOfTopLevelItem(item)
        wf = self._tree_widget._workflow
        step = wf[row]
        step.process_output = not step.process_output
        self.set_workflow(wf)
        self._modified = True

    def _find_window(self, item: QtW.QTreeWidgetItem) -> None:
        row = self._tree_widget.indexOfTopLevelItem(item)
        step = self._tree_widget._workflow[row]
        main = get_main_window(self)
        for i_tab, tab in main.tabs.enumerate():
            for i_win, win in tab.enumerate():
                if win._identifier == step.id:
                    main.tabs.current_index = i_tab
                    main.tabs[i_tab].current_index = i_win
                    return None
        raise ValueError("No window in the main window matches the workflow step.")

    def _execute_upto(self, item: QtW.QTreeWidgetItem) -> None:
        row = self._tree_widget.indexOfTopLevelItem(item)
        wf = self._tree_widget._workflow
        step = wf[row]
        wf.filter(step.id).compute(process_output=True)

    def _edit_user_input(self, item: QtW.QTreeWidgetItem) -> None:
        row = self._tree_widget.indexOfTopLevelItem(item)
        wf = self._tree_widget._workflow
        step = wf[row]
        if not isinstance(step, _wf.UserInput):
            raise ValueError(f"Expected UserInput, got {type(step)}")
        resp = request_values(
            label={"value": step.label, "label": "Label"},
            doc={"value": step.doc, "label": "Docstring"},
            title="Edit User Input",
            parent=self,
        )
        if resp:
            step.label = resp["label"]
            step.doc = resp["doc"]
            self.set_workflow(wf)  # update
            self._modified = True

    def _on_right_clicked(self, item: QtW.QTreeWidgetItem, pos: QtCore.QPoint) -> None:
        self._make_context_menu(item).exec(pos)

    def _make_context_menu(self, item: QtW.QTreeWidgetItem) -> QtW.QMenu:
        menu = QtW.QMenu(self)
        step = _step_for_item(item)
        a0 = menu.addAction(
            "Replace with file reader",
            lambda: self._replace_with_file_reader(item, "file"),
        )
        a0.setToolTip("Replace the selected item with a file reader")
        a1 = menu.addAction(
            "Replace with model input",
            lambda: self._replace_with_file_reader(item, "model"),
        )
        a1.setToolTip("Replace the selected item with a model input")
        a2 = menu.addAction("To be added", lambda: self._toggle_to_be_added(item))
        a2.setCheckable(True)
        a2.setChecked(step.process_output)
        a2.setToolTip(
            "Mark the selected item to add the output to the main window after the "
            "workflow execution (even if it's an intermediate step)"
        )
        menu.addSeparator()
        a3 = menu.addAction("Execute upto here", lambda: self._execute_upto(item))
        a3.setToolTip("Execute the workflow up to the selected item")
        a4 = menu.addAction("Find window", lambda: self._find_window(item))
        a4.setToolTip("Find the window that corresponds to the selected item")
        a5 = menu.addAction("Edit ...", lambda: self._edit_user_input(item))
        a5.setToolTip("Edit the selected user input item")

        # update enabled state
        a0.setEnabled(self.is_editable())
        a1.setEnabled(self.is_editable())
        a2.setEnabled(self.is_editable())
        a5.setEnabled(self.is_editable())
        if isinstance(step, _wf.UserInput):
            a3.setEnabled(False)
        else:
            a5.setEnabled(False)
        tree = self._tree_widget
        if tree.indexOfTopLevelItem(item) == tree.topLevelItemCount() - 1:
            a2.setEnabled(False)
        return menu


class QWorkflowTree(QtW.QTreeWidget):
    right_clicked = QtCore.Signal(QtW.QTreeWidgetItem, QtCore.QPoint)

    def __init__(self, workflow: _wf.Workflow):
        super().__init__()
        self._workflow = workflow
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.NoSelection)
        self._current_index = 0
        self._selection_color = QtGui.QColor("#777777")
        self._hover_color = QtGui.QColor(self._selection_color)
        self._selection_model = SelectionModel(
            row_count=lambda: len(self._workflow), col_count=lambda: 1
        )
        self._selection_model.moved.connect(self._on_moved)
        self._id_to_index_map = {}

    def set_workflow(self, workflow: _wf.Workflow):
        self.clear()
        self._workflow = workflow
        self._id_to_index_map = workflow.id_to_index_map()
        for each in workflow:
            item = _step_to_item(each)
            _add_common_child(item, each)
            self.addTopLevelItem(item)

    @QtCore.Property(QtGui.QColor)
    def selectionColor(self):
        return self._selection_color

    @selectionColor.setter
    def selectionColor(self, color: QtGui.QColor):
        self._selection_color = color
        self._hover_color = QtGui.QColor(color)
        self._hover_color.setAlpha(128)

    def keyPressEvent(self, event):
        _key = event.key()
        _mod = event.modifiers()
        if _key == QtCore.Qt.Key.Key_Up:
            if _mod & QtCore.Qt.KeyboardModifier.ControlModifier:
                dr = -99999999
            else:
                dr = -1
            self._selection_model.move(dr, 0)
            return
        elif _key == QtCore.Qt.Key.Key_Down:
            if _mod & QtCore.Qt.KeyboardModifier.ControlModifier:
                dr = 99999999
            else:
                dr = 1
            self._selection_model.move(dr, 0)
            return
        return super().keyPressEvent(event)

    def _on_moved(self, src: Index, dst: Index) -> None:
        """Update the view."""
        self.update()

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        """Paint table and the selection."""
        super().paintEvent(event)
        painter = QtGui.QPainter(self.viewport())

        idx = self._selection_model.current_index
        item = self.itemFromIndex(self.model().index(idx.row, 0))
        if item is None:
            return None
        rect = self._rect_for_row(idx.row)
        pen = QtGui.QPen(self._selection_color, 2)
        painter.setPen(pen)
        painter.drawRect(rect)

        # draw parents
        pen.setWidth(1)
        pen.setStyle(QtCore.Qt.PenStyle.DashLine)
        painter.setPen(pen)
        for step in self._workflow[idx.row].iter_parents():
            try:
                index = self._id_to_index_map[step]
            except KeyError:  # This should not happen but just in case
                continue
            rect = self._rect_for_row(index)
            painter.drawRect(rect)

        return None

    def _rect_for_row(self, row: int) -> QtCore.QRect:
        index = self.model().index(row, 0)
        item = self.ancestor_item(index)
        if item is None:
            return QtCore.QRect()
        if item.isExpanded():
            rect = self.visualRect(index)
            child = item.child(item.childCount() - 1)
            if child is not None:
                index = self.indexFromItem(child)
                rect = rect.united(self.visualRect(index))
        else:
            rect = self.visualRect(index)
        return rect.adjusted(1, 1, -1, -1)

    def ancestor_item(self, index: QtCore.QModelIndex) -> QtW.QTreeWidgetItem | None:
        item = self.itemFromIndex(index)
        if item is None:
            return None
        while parent := item.parent():
            item = parent
        return item

    def mousePressEvent(self, e: QtGui.QMouseEvent):
        index = self.indexAt(e.pos())
        if item := self.ancestor_item(index):
            if e.buttons() & QtCore.Qt.MouseButton.LeftButton:
                row = self.indexOfTopLevelItem(item)
                self._selection_model.jump_to(row, 0)
            elif e.buttons() & QtCore.Qt.MouseButton.RightButton:
                self.right_clicked.emit(item, self.mapToGlobal(e.pos()))
        return super().mousePressEvent(e)

    # drag-and-drop
    def mouseMoveEvent(self, e: QtGui.QMouseEvent):
        if e.buttons() & QtCore.Qt.MouseButton.LeftButton:
            if drag := self._make_drag(e.pos()):
                drag.exec()
            return None
        return super().mouseMoveEvent(e)

    def _make_drag(self, pos: QtCore.QPoint):
        if not self.indexAt(pos).isValid():
            return
        row = self._selection_model.current_index.row
        if 0 <= row < len(self._workflow):
            wf_filt = self._workflow.filter(self._workflow[row].id)
            return drag_model(
                WidgetDataModel(
                    value=wf_filt,
                    type=StandardType.WORKFLOW,
                    title="Subset of workflow",
                ),
                desc="workflow node",
                source=self.parent(),
                text_data=str(wf_filt),
                exec=False,
            )


_STEP_ROLE = QtCore.Qt.ItemDataRole.UserRole


@singledispatch
def _step_to_item(step: _wf.WorkflowStep) -> QtW.QTreeWidgetItem:
    raise ValueError(f"Unknown workflow node type {type(step)}")


@_step_to_item.register
def _(step: _wf.LocalReaderMethod) -> QtW.QTreeWidgetItem:
    o = _mark_for_to_be_added(step)
    item = QtW.QTreeWidgetItem([f"[{o}Local Path] (type={step.output_model_type!r})"])
    if isinstance(step.path, list):
        for i, path in enumerate(step.path):
            item.addChild(QtW.QTreeWidgetItem([f"({i}) {path.as_posix()}"]))
    else:
        item.addChild(QtW.QTreeWidgetItem([f"{step.path.as_posix()}"]))
    item.addChild(QtW.QTreeWidgetItem([f"plugin = {step.plugin!r}"]))
    item.setToolTip(0, str(step.path))
    item.setData(0, _STEP_ROLE, step)
    return item


@_step_to_item.register
def _(step: _wf.RemoteReaderMethod) -> QtW.QTreeWidgetItem:
    o = _mark_for_to_be_added(step)
    item = QtW.QTreeWidgetItem([f"[{o}Remote Path] (type={step.output_model_type!r})"])
    item.addChild(QtW.QTreeWidgetItem([f"{step.to_str()}"]))
    item.addChild(QtW.QTreeWidgetItem([f"plugin = {step.plugin!r}"]))
    item.setToolTip(0, str(step.path))
    item.setData(0, _STEP_ROLE, step)
    return item


@_step_to_item.register
def _(step: _wf.UserModification) -> QtW.QTreeWidgetItem:
    item = QtW.QTreeWidgetItem(["[User Modification]"])
    item.setData(0, _STEP_ROLE, step)
    return item


@_step_to_item.register
def _(step: _wf.UserInput) -> QtW.QTreeWidgetItem:
    o = _mark_for_to_be_added(step)
    if step.how == "file":
        text = "File Input"
    elif step.how == "model":
        text = "Model Input"
    else:
        text = "User Input"
    item = QtW.QTreeWidgetItem([f"[{o}{text}]"])
    item.addChild(QtW.QTreeWidgetItem([f"label = {step.label!r}"]))
    item.addChild(QtW.QTreeWidgetItem([f"doc = {step.doc!r}"]))
    item.setToolTip(0, f"{text}\nlabel = {step.label}\ndoc = {step.doc}")
    item.setData(0, _STEP_ROLE, step)
    return item


@_step_to_item.register
def _(step: _wf.CommandExecution) -> QtW.QTreeWidgetItem:
    o = _mark_for_to_be_added(step)
    item = QtW.QTreeWidgetItem([f"[{o}Command] {step.command_id}"])
    item.setToolTip(0, step.command_id)
    for param in step.parameters or []:
        if isinstance(param, _wf.UserParameter):
            child = QtW.QTreeWidgetItem([f"(parameter) {param.name} = {param.value!r}"])
        elif isinstance(param, _wf.ModelParameter):
            child = QtW.QTreeWidgetItem(
                [f"(parameter) {param.name} = <data model, type={param.model_type!r}>"]
            )
        elif isinstance(param, _wf.WindowParameter):
            child = QtW.QTreeWidgetItem(
                [f"(parameter) {param.name} = <window, type={param.model_type!r}>"]
            )
        elif isinstance(param, _wf.ListOfModelParameter):
            short_desc = "<subwindows>" if param.is_window else "<models>"
            child = QtW.QTreeWidgetItem([f"(parameter) {param.name} = {short_desc}"])
        else:
            raise ValueError(f"Unknown parameter type {type(param)}")
        item.addChild(child)
    for ctx in step.contexts:
        if isinstance(ctx, _wf.ModelParameter):
            child = QtW.QTreeWidgetItem(
                [f"(context) <data model, type={ctx.model_type!r}>"]
            )
        elif isinstance(ctx, _wf.WindowParameter):
            child = QtW.QTreeWidgetItem(
                [f"(context) <window, type={ctx.model_type!r}>"]
            )
        else:
            raise ValueError(f"Unknown context type {type(ctx)}")
        item.addChild(child)
    if (dt := step.execution_time) > 0.0:
        item.addChild(
            QtW.QTreeWidgetItem([f"execution time = {timedelta(seconds=dt)}"])
        )
    item.setData(0, _STEP_ROLE, step)
    return item


@_step_to_item.register
def _(step: _wf.ProgrammaticMethod) -> QtW.QTreeWidgetItem:
    item = QtW.QTreeWidgetItem(
        [f"[Programmatic Method] (type={step.output_model_type!r})"]
    )
    item.setData(0, _STEP_ROLE, step)
    return item


def _add_common_child(item: QtW.QTreeWidgetItem, step: _wf.WorkflowStep):
    item.addChild(
        QtW.QTreeWidgetItem([f"datetime = {step.datetime:%Y-%m-%d %H:%M:%S}"])
    )
    return item


def _mark_for_to_be_added(step: _wf.WorkflowStep) -> str:
    return "● " if step.process_output else ""


def _step_for_item(item: QtW.QTreeWidgetItem) -> _wf.WorkflowStep:
    step = item.data(0, _STEP_ROLE)
    return step



from __future__ import annotations
from datetime import datetime
from pathlib import Path
import warnings
import shutil
from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtCore, QtGui
from himena import _drag
from himena_builtins.qt.explorer._widget_ssh import (
    QSSHRemoteExplorerWidget,
    make_paste_remote_files_worker,
)

if TYPE_CHECKING:
    from himena.qt import MainWindowQt
    from himena_builtins.qt.explorer import FileExplorerConfig


class QFileSystemModel(QtW.QFileSystemModel):
    def __init__(self):
        super().__init__()
        self.setRootPath(Path.cwd().as_posix())

    def columnCount(self, parent) -> int:
        return 1

    def data(self, index: QtCore.QModelIndex, role: int):
        if role == QtCore.Qt.ItemDataRole.SizeHintRole:
            return QtCore.QSize(18, 16)
        return super().data(index, role)

    def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlag:
        # NOTE: renaming of item triggers renaming of the file by default.
        return super().flags(index) | QtCore.Qt.ItemFlag.ItemIsEditable


class QRootPathEdit(QtW.QWidget):
    """Widget to specify the root path of the file explorer."""

    rootChanged = QtCore.Signal(Path)

    def __init__(self) -> None:
        super().__init__()
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self._path_edit = QtW.QLabel()
        self._path_edit.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
        )
        self._btn_set_root = QtW.QPushButton("...")
        self._btn_set_root.setFixedWidth(20)
        self._btn_set_root.clicked.connect(self._select_root_path)
        layout.addWidget(self._path_edit)
        layout.addWidget(self._btn_set_root)

    def _select_root_path(self):
        path = QtW.QFileDialog.getExistingDirectory(self, "Select Root Path")
        if not path:
            return
        self.set_root_path(path)

    def set_root_path(self, path: str | Path):
        path = Path(path)
        self._path_edit.setText("/" + path.name)
        self.rootChanged.emit(path)
        self._path_edit.setToolTip(path.as_posix())


class QExplorerWidget(QtW.QWidget):
    """A normal file explorer widget."""

    open_file_requested = QtCore.Signal(Path)

    def __init__(self, ui: MainWindowQt) -> None:
        super().__init__()
        self._ui = ui
        self._root = QRootPathEdit()
        self._file_tree = QFileTree(ui)
        layout = QtW.QVBoxLayout(self)
        layout.addWidget(self._root)
        layout.addWidget(self._file_tree)
        self._root.set_root_path(Path.cwd())
        self._root.rootChanged.connect(self._file_tree.setRootPath)
        self._file_tree.fileDoubleClicked.connect(self.open_file_requested.emit)
        self.open_file_requested.connect(ui.read_file)

    def update_configs(self, cfg: FileExplorerConfig):
        self._file_tree._config = cfg


class QExtendedFileSystemModel(QFileSystemModel):
    def data(self, index, role):
        if role == QtCore.Qt.ItemDataRole.ToolTipRole:
            return self._tooltip_for_index(index)
        return super().data(index, role)

    def _tooltip_for_index(self, index: QtCore.QModelIndex) -> str:
        if not index.isValid():
            return ""
        path = Path(self.filePath(index))
        if path.exists():
            stat = path.stat()
            size_human_readable = QtCore.QLocale().formattedDataSize(stat.st_size, 2)
            time_last_modified = datetime.fromtimestamp(stat.st_mtime)
            return (
                f"{path.as_posix()}\n"
                f"Size: {size_human_readable}\n"
                f"Last modified: {time_last_modified:%Y-%m-%d %H:%M:%S}"
            )
        else:
            return "<Deleted>"


class QFileTree(QtW.QTreeView):
    fileDoubleClicked = QtCore.Signal(Path)

    def __init__(self, ui: MainWindowQt) -> None:
        from himena_builtins.qt.explorer import FileExplorerConfig

        super().__init__()
        self._ui = ui
        self._model = QExtendedFileSystemModel()
        self.setHeaderHidden(True)
        self.setEditTriggers(QtW.QAbstractItemView.EditTrigger.EditKeyPressed)
        self.setTextElideMode(QtCore.Qt.TextElideMode.ElideNone)
        self.setModel(self._model)
        self.setRootIndex(self._model.index(self._model.rootPath()))
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.doubleClicked.connect(self._double_clicked)

        # context menu
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_context_menu)

        self.setAcceptDrops(True)
        self._config = FileExplorerConfig()

    def setRootPath(self, path: Path):
        path = Path(path)
        self._model.setRootPath(path.as_posix())
        self.setRootIndex(self._model.index(path.as_posix()))

    def _make_context_menu(self, index: QtCore.QModelIndex):
        menu = QtW.QMenu(self)
        path = Path(self._model.filePath(index))
        menu.addAction("Open", lambda: self._ui.read_file(path))
        menu.addSeparator()
        menu.addAction("Copy", lambda: self._ui.set_clipboard(files=[path]))
        menu.addAction("Copy Path", lambda: self._ui.set_clipboard(text=str(path)))
        menu.addAction("Paste", self._paste_from_clipboard)
        menu.addSeparator()
        menu.addAction("Rename", lambda: self.edit(index))
        selected_indices = self.selectedIndexes()
        menu.addAction("Delete", lambda: self._move_to_trash(selected_indices))
        return menu

    def _show_context_menu(self, pos: QtCore.QPoint):
        index = self.indexAt(pos)
        if not index.isValid():
            return
        menu = self._make_context_menu(index)
        menu.exec(self.viewport().mapToGlobal(pos))

    def _double_clicked(self, index: QtCore.QModelIndex):
        idx = self._model.index(index.row(), 0, index.parent())
        path = Path(self._model.filePath(idx))
        if path.is_dir():
            return
        self.fileDoubleClicked.emit(path)
        return None

    # drag-and-drop
    def mouseMoveEvent(self, e: QtGui.QMouseEvent):
        if e.buttons() & QtCore.Qt.MouseButton.LeftButton:
            self._start_drag(e.pos())
            return None
        return super().mouseMoveEvent(e)

    def _make_drag(self) -> QtGui.QDrag:
        mime = QtCore.QMimeData()
        selected_indices = self.selectedIndexes()
        urls = [self._model.filePath(idx) for idx in selected_indices]
        mime.setUrls([QtCore.QUrl.fromLocalFile(url) for url in urls])
        drag = QtGui.QDrag(self)
        drag.setMimeData(mime)
        if (nfiles := len(selected_indices)) == 1:
            pixmap = self._model.fileIcon(selected_indices[0]).pixmap(10, 10)
        else:
            qlabel = QtW.QLabel(f"{nfiles} files")
            pixmap = QtGui.QPixmap(qlabel.size())
            qlabel.render(pixmap)
        drag.setPixmap(pixmap)
        cursor = QtGui.QCursor(QtCore.Qt.CursorShape.OpenHandCursor)
        drag.setDragCursor(cursor.pixmap(), QtCore.Qt.DropAction.MoveAction)
        return drag

    def _start_drag(self, pos: QtCore.QPoint):
        drag = self._make_drag()
        drag.exec(QtCore.Qt.DropAction.MoveAction)

    def dragEnterEvent(self, a0: QtGui.QDragEnterEvent):
        mime = a0.mimeData()
        if mime and (
            mime.hasUrls() or isinstance(mime.parent(), QSSHRemoteExplorerWidget)
        ):
            a0.accept()
        elif _drag.get_dragging_model() is not None:
            a0.accept()
        else:
            a0.ignore()

    def dragMoveEvent(self, a0: QtGui.QDragMoveEvent):
        mime = a0.mimeData()
        if mime and (
            mime.hasUrls() or isinstance(mime.parent(), QSSHRemoteExplorerWidget)
        ):
            a0.accept()
        elif _drag.get_dragging_model() is not None:
            a0.accept()
        else:
            a0.ignore()
            return
        index = self._get_directory_index(self.indexAt(a0.pos()))
        self.selectionModel().setCurrentIndex(
            index, QtCore.QItemSelectionModel.SelectionFlag.ClearAndSelect
        )
        a0.acceptProposedAction()

    def dropEvent(self, a0: QtGui.QDropEvent):
        dirpath = self._directory_for_index(self.indexAt(a0.pos()))
        if drag_model := _drag.drop():
            if self._config.allow_drop_data_to_save:
                data_model = drag_model.data_model()
                data_model.write_to_directory(dirpath)
        elif mime := a0.mimeData():
            self._paste_mime_data(mime, dirpath)

    def dragLeaveEvent(self, e):
        return super().dragLeaveEvent(e)

    def keyPressEvent(self, event):
        if event.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier:
            if event.key() == QtCore.Qt.Key.Key_C:
                self._ui.set_clipboard(
                    files=[
                        Path(self._model.filePath(idx))
                        for idx in self.selectedIndexes()
                    ]
                )
                return
            elif event.key() == QtCore.Qt.Key.Key_V:
                return self._paste_from_clipboard()
        return super().keyPressEvent(event)

    def _move_to_trash(self, indices: list[QtCore.QModelIndex]):
        ans = self._ui.exec_choose_one_dialog(
            "Move to Trash",
            "Are you sure you want to move the selected files to trash?",
            choices=["Delete", "Cancel"],
        )
        if ans == "Delete":
            for index in indices:
                QtCore.QFile.moveToTrash(self._model.filePath(index))

    def _paste_from_clipboard(self):
        if clipboard := QtW.QApplication.clipboard():
            dirpath = self._directory_for_index(self.currentIndex())
            self._paste_mime_data(clipboard.mimeData(), dirpath)

    def _paste_mime_data(self, mime: QtCore.QMimeData, dirpath: Path):
        ui = self._ui._backend_main_window
        if isinstance(par := mime.parent(), QSSHRemoteExplorerWidget):
            readers = par.readers_from_mime(mime)
            worker = make_paste_remote_files_worker(readers, dirpath)
            ui._job_stack.add_worker(worker, "Pasting remote files", total=len(readers))
            worker.start()
        else:
            self._paste_file(
                self._ui.clipboard.files,
                dirpath=dirpath,
                is_copy=True,
            )

    def _paste_file(self, paths: list[Path], dirpath: Path, is_copy: bool):
        dst_exists: list[Path] = []
        src_dst_set: list[tuple[Path, Path]] = []
        for src in paths:
            if not src.exists():
                warnings.warn(f"Path {src} does not exist.", stacklevel=2)
                continue
            dst = dirpath / src.name
            if src != dst:
                if dst.exists():
                    dst_exists.append(dst)
                src_dst_set.append((src, dst))
        if src_dst_set and self._config.allow_drop_file_to_move:
            if dst_exists:
                conflicts = "\n - ".join(p.name for p in dst_exists)
                answer = self._ui.exec_choose_one_dialog(
                    "Replace existing files?",
                    f"Name conflict in the destinations:\n{conflicts}",
                    ["Replace", "Skip", "Cancel"],
                )
                if answer == "Cancel":
                    return
                elif answer == "Replace":
                    pass
                else:
                    src_dst_set = [
                        (src, dst) for src, dst in src_dst_set if dst not in dst_exists
                    ]
            for src, dst in src_dst_set:
                if is_copy:
                    shutil.copy(src, dst)
                else:
                    shutil.move(src, dst)

    def _get_directory_index(self, index: QtCore.QModelIndex):
        if not index.isValid():
            return self.rootIndex()
        _is_directory = self.model().hasChildren(index)
        if _is_directory:
            return index
        else:
            return self.model().parent(index)

    def _directory_for_index(self, index: QtCore.QModelIndex) -> Path:
        dir_index = self._get_directory_index(index)
        if dirpath := self.model().filePath(dir_index):
            return Path(dirpath)
        else:
            raise ValueError(f"Invalid destination: {dirpath}")



from himena.workflow._base import WorkflowStep
from himena.workflow._graph import Workflow, compute, WorkflowStepType
from himena.workflow._caller import as_function, as_function_from_path
from himena.workflow._command import (
    CommandExecution,
    ListOfModelParameter,
    ModelParameter,
    UserModification,
    UserParameter,
    WindowParameter,
    parse_parameter,
)
from himena.workflow._reader import (
    LocalReaderMethod,
    ProgrammaticMethod,
    ReaderMethod,
    RemoteReaderMethod,
    UserInput,
)

__all__ = [
    "WorkflowStep",
    "Workflow",
    "compute",
    "as_function",
    "as_function_from_path",
    "WorkflowStepType",
    "ProgrammaticMethod",
    "ReaderMethod",
    "LocalReaderMethod",
    "RemoteReaderMethod",
    "UserInput",
    "CommandExecution",
    "parse_parameter",
    "ModelParameter",
    "UserModification",
    "WindowParameter",
    "UserParameter",
    "ListOfModelParameter",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Sequence
from pathlib import Path
import warnings
import yaml
import zipfile
import tempfile
from himena.session._session import AppSession, TabSession
from himena.session._utils import (
    write_model_by_title,
    write_metadata_by_title,
    replace_invalid_characters,
)
from himena.standards import read_metadata
from himena.widgets._wrapper import ParametricWindow
from himena.workflow import LocalReaderMethod
from himena.workflow._command import CommandExecution

if TYPE_CHECKING:
    from uuid import UUID
    from himena.widgets import MainWindow, TabArea

_SESSION_YAML = "session.yaml"


def update_from_directory(ui: MainWindow, path: str | Path) -> None:
    """Update GUI from a session directory."""
    dirpath = Path(path)
    with dirpath.joinpath(_SESSION_YAML).open("r") as f:
        yml = yaml.load(f, Loader=yaml.Loader)
    if not (isinstance(yml, dict) and "session" in yml):
        raise ValueError("Invalid session file.")
    wf_overrides = {}
    if yml.pop("session") == "main":
        session = AppSession.model_validate(yml)
        for tab_dir in dirpath.iterdir():
            if tab_dir.is_file():
                continue
            ith = int(tab_dir.stem.split("_")[0])
            for uuid, meth in _iter_reader_method(tab_dir, session.tabs[ith]):
                wf_overrides[uuid] = meth
        session.update_gui(ui, workflow_override=wf_overrides, dirpath=dirpath)
    else:
        session = TabSession.model_validate(yml)
        wf_overrides = dict(_iter_reader_method(dirpath, session))
        session.update_gui(ui, workflow_override=wf_overrides, dirpath=dirpath)
    return None


def _iter_reader_method(
    dirpath: Path,
    tab_session: TabSession,
) -> Iterator[tuple[UUID, LocalReaderMethod]]:
    for file in dirpath.iterdir():
        if file.suffix == ".himena-meta":
            continue
        ith_win = int(file.stem.rsplit("_")[0])
        win_sess = tab_session.windows[ith_win]
        uuid = win_sess.id
        meta_path = file.with_name(f"{ith_win}_{win_sess.title}.himena-meta")
        metadata = None
        if meta_path.exists():
            try:
                metadata = read_metadata(meta_path)
            except Exception as e:
                warnings.warn(
                    f"Failed to read metadata from {meta_path}: {e}",
                    RuntimeWarning,
                    stacklevel=2,
                )
        meth = LocalReaderMethod(path=file, metadata_override=metadata, id=uuid)
        yield uuid, meth.construct_workflow()


def update_from_zip(ui: MainWindow, path: str | Path) -> None:
    with (
        zipfile.ZipFile(path) as z,
        tempfile.TemporaryDirectory() as tmpdir,
    ):
        tmpdir = Path(tmpdir)
        z.extractall(tmpdir)
        update_from_directory(ui, tmpdir)
    return None


def dump_tab_to_directory(
    tab: TabArea,
    path: str | Path,
    save_copies: bool = False,
    allow_calculate: Sequence[str] = (),
):
    path = Path(path)
    path.mkdir(exist_ok=True)
    session = TabSession.from_gui(tab, allow_calculate=True)
    cmd_id_allowed = set(allow_calculate)
    with open(path / _SESSION_YAML, "w") as f:
        js = session.model_dump(mode="json")
        js = {"session": "tab", **js}
        yaml.dump(js, f, sort_keys=False)
    _dump_tab_to_directory_impl(tab, path, save_copies, cmd_id_allowed)


def dump_tab_to_zip(
    tab: TabArea,
    path: str | Path,
    save_copies: bool = False,
    allow_calculate: Sequence[str] = (),
):
    path = Path(path)
    with tempfile.TemporaryDirectory() as tmpdir, zipfile.ZipFile(path, "w") as z:
        tmpdir = Path(tmpdir)
        dump_tab_to_directory(
            tab, tmpdir, save_copies=save_copies, allow_calculate=allow_calculate
        )
        for file in tmpdir.rglob("*"):
            z.write(file, file.relative_to(tmpdir))
    return None


def _dump_tab_to_directory_impl(
    tab: TabArea,
    dirname: Path,
    save_copies: bool,
    cmd_id_allowed: set[str],
):
    for i, win in enumerate(tab):
        if isinstance(win, ParametricWindow):
            continue
        prefix = str(i)
        read_from = win._determine_read_from()
        if read_from is not None and isinstance(read_from[0], Path) and not save_copies:
            write_metadata_by_title(win, dirname, prefix)
        elif (
            isinstance(step := win._widget_workflow.last(), CommandExecution)
            and step.command_id in cmd_id_allowed
        ):
            write_metadata_by_title(win, dirname, prefix)
        elif win.supports_to_model and win.supports_update_model:
            write_model_by_title(win, dirname, prefix=prefix)
        else:
            # Cannot read or write the window state, must rely on command execution
            write_metadata_by_title(win, dirname, prefix)


def dump_directory(
    ui: MainWindow,
    path: str | Path,
    save_copies: bool = False,
    allow_calculate: Sequence[str] = (),
):
    """Dump the main window state as a directory.

    The content will be something like:
    ```
    my_session_directory/
      ├── session.yaml
      └── 0_tab_name/
            ├── 00_xxx.txt
            ├── 01_yyy.csv
            :
    ```
    """
    path = Path(path)
    path.mkdir(exist_ok=True)
    session = AppSession.from_gui(ui, allow_calculate=True)
    cmd_id_allowed = set(allow_calculate)
    with open(path / _SESSION_YAML, "w") as f:
        js = session.model_dump(mode="json")
        js = {"session": "main", **js}
        yaml.dump(js, f, sort_keys=False)
    for i_tab, tab in enumerate(ui.tabs):
        tab_title = replace_invalid_characters(tab.title)
        dirname = path / f"{i_tab}_{tab_title}"
        dirname.mkdir()
        _dump_tab_to_directory_impl(tab, dirname, save_copies, cmd_id_allowed)
    return None


def dump_zip(
    ui: MainWindow,
    path: str | Path,
    save_copies: bool = False,
    allow_calculate: Sequence[str] = (),
):
    """Dump the main window state as a stand-alone zip file.

    The content will be something like:
    ```
    my.session.zip/
      ├── session.yaml
      └── 0_tab_name/
            ├── 00_xxx.txt
            ├── 01_yyy.csv
            :
    ```
    """
    path = Path(path)
    with tempfile.TemporaryDirectory() as tmpdir, zipfile.ZipFile(path, "w") as z:
        tmpdir = Path(tmpdir)
        dump_directory(
            ui, tmpdir, save_copies=save_copies, allow_calculate=allow_calculate
        )
        for file in tmpdir.rglob("*"):
            z.write(file, file.relative_to(tmpdir))
    return None



from importlib import import_module
import sys
from himena.consts import MenuId
from himena.widgets import MainWindow
from himena.types import ClipboardDataModel
from himena._app_model.actions._registry import ACTIONS


def _open_url(url: str):
    import webbrowser

    webbrowser.open(url)


REPOSITORY_URL = "https://github.com/hanjinliu/himena"
DOCUMENTATION_URL = "https://hanjinliu.github.io/himena/"


@ACTIONS.append_from_fn(
    id="github-repo",
    title="GitHub Repository",
    menus=[{"id": MenuId.HELP, "group": "98_open-site-group"}],
)
def open_github_repo() -> None:
    """Open the GitHub repository in the default browser."""
    _open_url(REPOSITORY_URL)


@ACTIONS.append_from_fn(
    id="documentation",
    title="Documentation",
    menus=[{"id": MenuId.HELP, "group": "98_open-site-group"}],
)
def open_documentation() -> None:
    """Open the himena documentation."""
    _open_url(DOCUMENTATION_URL)


@ACTIONS.append_from_fn(
    id="report-issue",
    title="Report Issue",
    menus=[{"id": MenuId.HELP, "group": "98_open-site-group"}],
)
def report_issue() -> None:
    """Open the issue tracker in the default browser."""
    _open_url(REPOSITORY_URL + "/issues/new")


@ACTIONS.append_from_fn(
    id="show-about",
    title="About",
    menus=[{"id": MenuId.HELP, "group": "99_about-group"}],
)
def show_about(ui: MainWindow) -> None:
    """Show the about dialog."""
    from himena import __version__

    lines: list[str] = []
    lines.append(f"<h3>himena {__version__}</h3>")
    lines.append("<b>Dependencies:</b>")
    for dependency in [
        "numpy",
        "app_model",
        "pydantic",
        "superqt",
        "magicgui",
        "platformdirs",
        "pygments",
        "pyyaml",
        "pillow",
        "tabulate",
    ]:
        try:
            mod = import_module(dependency)
            lines.append(f"&nbsp;&nbsp;{dependency}={mod.__version__}")
        except ImportError:
            pass

    lines.append("")
    lines.append("<b>Python:</b>")
    lines.append(f"&nbsp;&nbsp;{sys.version}")
    lines.append("<b>Platform:</b>")
    lines.append(f"&nbsp;&nbsp;{sys.platform}")
    info = "<br>".join(lines)
    if ui.exec_choose_one_dialog(
        "About",
        message=info,
        choices=[("Copy", True), ("OK", False)],
        how="buttons",
    ):
        ui.clipboard = ClipboardDataModel(html=info)
    return None



from himena._cli.namespace import HimenaCliNamespace


def _assert_profile_not_none(profile: str | None) -> None:
    if profile is None:
        raise ValueError("Profile name is required with the --new option.")


def remove_profile(args: HimenaCliNamespace):
    _assert_profile_not_none(args.remove)
    args.assert_args_not_given()
    from himena.profile import remove_app_profile

    remove_app_profile(args.remove)
    print(f"Profile {args.remove!r} is removed.")


def new_profile(args: HimenaCliNamespace):
    _assert_profile_not_none(args.new)
    args.assert_args_not_given()
    from himena.profile import new_app_profile

    new_app_profile(args.new)
    print(
        f"Profile {args.new!r} is created. You can start the application with:\n"
        f"$ himena {args.new}"
    )



from typing import Callable
import numpy as np
from numpy.testing import assert_array_equal
from himena import MainWindow, StandardType

def test_convert_text(make_himena_ui: Callable[..., MainWindow]):
    himena_ui = make_himena_ui("mock")
    win = himena_ui.add_object("a,b,c\n1,2,3", type=StandardType.TEXT)
    himena_ui.exec_action("builtins:text-to-table")
    assert himena_ui.current_model.type == StandardType.TABLE
    assert himena_ui.current_model.value.tolist() == [["a", "b", "c"], ["1", "2", "3"]]

    himena_ui.add_object("1.2\n4.2\n", type=StandardType.TEXT)
    himena_ui.exec_action("builtins:text-to-table")
    assert himena_ui.current_model.type == StandardType.TABLE
    assert himena_ui.current_model.value.tolist() == [["1.2"], ["4.2"]]

    himena_ui.add_object("1\t5\n4\t2\n", type=StandardType.TEXT)
    himena_ui.exec_action("builtins:text-to-table")
    assert himena_ui.current_model.type == StandardType.TABLE
    assert himena_ui.current_model.value.tolist() == [["1", "5"], ["4", "2"]]

    himena_ui.current_window = win
    himena_ui.exec_action("builtins:text-to-dataframe")
    assert himena_ui.current_model.type == StandardType.DATAFRAME

    win = himena_ui.add_object("-1,3,4.3\n1.2,2,3", type=StandardType.TEXT)
    himena_ui.exec_action("builtins:text-to-array")
    assert_array_equal(himena_ui.current_model.value, np.array([[-1, 3, 4.3], [1.2, 2, 3]]))

    win = himena_ui.add_object("<b>a</b><br>3", type=StandardType.HTML)
    himena_ui.exec_action("builtins:to-plain-text")
    # FIXME: return value of toHtml() contains "\n". How to process them?
    # assert himena_ui.current_model.type == StandardType.TEXT
    # assert himena_ui.current_model.value == "a\n3"

def test_convert_table_to_dataframe(make_himena_ui: Callable[..., MainWindow]):
    himena_ui = make_himena_ui("mock")
    himena_ui.add_object([["xx", "yyy"], ["e", -0.33], ["f", 2.3]], type=StandardType.TABLE)
    himena_ui.exec_action("builtins:table-to-dataframe")
    assert himena_ui.current_model.type == StandardType.DATAFRAME
    assert isinstance(himena_ui.current_model.value["xx"].dtype, np.dtypes.StringDType)
    assert himena_ui.current_model.value["yyy"].dtype == np.float64

    himena_ui.add_object([["p", "q"], ["4", ""], ["", "-0.3"]], type=StandardType.TABLE)
    himena_ui.exec_action("builtins:table-to-dataframe")
    assert himena_ui.current_model.type == StandardType.DATAFRAME
    assert himena_ui.current_model.value["p"].dtype == np.float64
    assert himena_ui.current_model.value["q"].dtype == np.float64

def test_convert_others(make_himena_ui: Callable[..., MainWindow]):
    himena_ui = make_himena_ui("mock")
    win = himena_ui.add_object(
        {"a": [1, 2], "value": ["p", "q"]}, type=StandardType.DATAFRAME,
    )
    himena_ui.exec_action("builtins:dataframe-to-table")
    assert himena_ui.current_model.type == StandardType.TABLE
    assert himena_ui.current_model.value.tolist() == [["a", "value"], ["1", "p"], ["2", "q"]]

    win = himena_ui.add_object({"a": [1, 2], "bbb": ["p", "q"]}, type=StandardType.DATAFRAME)
    himena_ui.exec_action("builtins:dataframe-to-text", with_params={"format": "CSV"})
    himena_ui.current_window = win
    himena_ui.exec_action("builtins:dataframe-to-text", with_params={"format": "TSV"})
    himena_ui.current_window = win
    himena_ui.exec_action("builtins:dataframe-to-text", with_params={"format": "Markdown"})
    himena_ui.current_window = win
    himena_ui.exec_action("builtins:dataframe-to-text", with_params={"format": "Latex"})
    himena_ui.current_window = win
    himena_ui.exec_action("builtins:dataframe-to-text", with_params={"format": "rST"})
    himena_ui.current_window = win
    himena_ui.exec_action("builtins:dataframe-to-text", with_params={"format": "HTML"})

    himena_ui.add_object([[1, 2], [-1, 2], [3, 2]], type=StandardType.TABLE)
    himena_ui.exec_action("builtins:table-to-array")

    himena_ui.add_object({"x": [4, 3], "y": [3.1, 2.3]}, type=StandardType.DATAFRAME)
    himena_ui.exec_action("builtins:dataframe-to-image-rois", with_params={"roi_type": "point"})
    himena_ui.add_object({"x": [4, 3], "y": [3.1, 2.3], "width": [2, 3], "height": [5, 6]}, type=StandardType.DATAFRAME)
    himena_ui.exec_action("builtins:dataframe-to-image-rois", with_params={"roi_type": "rectangle"})
    himena_ui.add_object({"x": [1, 2, 3], "y": [5, 3, 4]}, type=StandardType.DATAFRAME)
    himena_ui.exec_action("builtins:dataframe-to-dataframe-plot")
    himena_ui.add_object(np.arange(6).reshape(2, 3), type=StandardType.ARRAY)
    himena_ui.exec_action("builtins:array-to-table")



from io import StringIO
from typing import TYPE_CHECKING
import numpy as np
from himena.plugins import register_function, configure_gui, configure_submenu
from himena.types import ClipboardDataModel, Parametric, WidgetDataModel
from himena.standards.model_meta import TableMeta
from himena.consts import StandardType, MenuId
from himena.utils.misc import table_to_text
from himena.widgets import SubWindow

if TYPE_CHECKING:
    from himena_builtins.qt.widgets.table import QSpreadsheet

configure_submenu(MenuId.TOOLS_TABLE, group="20_builtins", order=1)


@register_function(
    title="Crop selection",
    types=StandardType.TABLE,
    menus=[MenuId.TOOLS_TABLE],
    command_id="builtins:table:crop",
)
def crop_selection(model: WidgetDataModel["np.ndarray"]) -> Parametric:
    """Crop the table data at the selection."""

    @configure_gui(selection={"bind": lambda *_: _get_selection(model)})
    def run_crop_selection(
        selection: tuple[tuple[int, int], tuple[int, int]],
    ) -> WidgetDataModel:
        (r0, r1), (c0, c1) = selection
        arr_str = model.value
        arr_new = arr_str[r0:r1, c0:c1]
        out = model.with_value(arr_new)
        if isinstance(meta := out.metadata, TableMeta):
            meta.selections = []
        return out

    return run_crop_selection


@register_function(
    title="Change separator ...",
    types=StandardType.TABLE,
    menus=[MenuId.TOOLS_TABLE],
    command_id="builtins:table:change-separator",
)
def change_separator(model: WidgetDataModel) -> Parametric:
    """Change the separator of the table data."""
    arr_str = model.value
    meta = _cast_meta(model.metadata)
    sep = meta.separator
    if sep is None:
        raise ValueError("Current separator of the table is unknown.")

    def change_separator(separator: str = ",") -> None:
        buf = _arr_to_buf(arr_str, sep)
        arr_new = np.loadtxt(
            buf,
            delimiter=separator.encode().decode("unicode_escape"),
            dtype=np.dtypes.StringDType(),
        )
        meta = _cast_meta(model.metadata)
        meta.separator = separator
        return model.with_value(arr_new, metadata=meta, update_inplace=True)

    return change_separator


@register_function(
    title="Copy as CSV",
    types=StandardType.TABLE,
    menus=[MenuId.TOOLS_TABLE_COPY, "/model_menu/copy"],
    command_id="builtins:table:copy-as-csv",
)
def copy_as_csv(model: WidgetDataModel) -> ClipboardDataModel:
    """Copy the table data as CSV format."""
    return _to_clipboard_data_model(model, "CSV")


@register_function(
    title="Copy as Markdown",
    types=StandardType.TABLE,
    menus=[MenuId.TOOLS_TABLE_COPY, "/model_menu/copy"],
    command_id="builtins:table:copy-as-markdown",
)
def copy_as_markdown(model: WidgetDataModel) -> ClipboardDataModel:
    """Copy the table data as Markdown format."""
    return _to_clipboard_data_model(model, "Markdown")


@register_function(
    title="Copy as HTML",
    types=StandardType.TABLE,
    menus=[MenuId.TOOLS_TABLE_COPY, "/model_menu/copy"],
    command_id="builtins:table:copy-as-html",
)
def copy_as_html(model: WidgetDataModel) -> ClipboardDataModel:
    """Copy the table data as HTML format."""
    return _to_clipboard_data_model(model, "HTML")


@register_function(
    title="Copy as rST",
    types=StandardType.TABLE,
    menus=[MenuId.TOOLS_TABLE_COPY, "/model_menu/copy"],
    command_id="builtins:table:copy-as-rst",
)
def copy_as_rst(model: WidgetDataModel) -> ClipboardDataModel:
    """Copy the table data as reStructuredText format."""
    return _to_clipboard_data_model(model, "rST")


@register_function(
    title="Insert incrementing numbers",
    types=StandardType.TABLE,
    menus=[MenuId.TOOLS_TABLE],
    command_id="builtins:table:insert-incrementing-numbers",
)
def insert_incrementing_numbers(win: SubWindow["QSpreadsheet"]) -> Parametric:
    """Insert incrementing numbers (0, 1, 2, ...) in-place to the selected range."""
    widget = win.widget

    def _get_selection(*_):
        meta = _cast_meta(win.to_model().metadata)
        sels = meta.selections
        if sels is None or len(sels) != 1:
            raise ValueError("Table must contain single selection to crop.")
        return sels[0]

    @configure_gui(title="Change separator", selection={"bind": _get_selection})
    def run_insert(
        selection: tuple[tuple[int, int], tuple[int, int]],
        start: int = 0,
        step: int = 1,
    ) -> None:
        (r0, r1), (c0, c1) = selection
        length = (r1 - r0) * (c1 - c0)
        values = [str(i) for i in range(start, start + length * step, step)]
        if r1 - r0 != 1 and c1 - c0 != 1:
            raise ValueError("Select a single row or column.")
        nr, nc = widget.model()._arr.shape
        if nr < r1 or nc < c1:
            widget.array_expand(r1, c1)
        target = widget.model()._arr
        if r1 - r0 == 1:
            target[r0:r1, c0:c1] = np.array(values, dtype=target.dtype).reshape(1, -1)
        else:
            target[r0:r1, c0:c1] = np.array(values, dtype=target.dtype).reshape(-1, 1)
        return

    return run_insert


def _cast_meta(meta) -> TableMeta:
    if not isinstance(meta, TableMeta):
        raise ValueError("Table must have a TableMeta as the metadata")
    return meta


def _arr_to_buf(arr: "np.ndarray", sep: str = ",") -> StringIO:
    buf = StringIO()
    np.savetxt(buf, arr, fmt="%s", delimiter=sep)
    buf.seek(0)
    return buf


def _get_selection(
    model: WidgetDataModel["np.ndarray"],
    allow_no_selection: bool = False,
) -> tuple[tuple[int, int], tuple[int, int]]:
    meta = _cast_meta(model.metadata)
    sels = meta.selections
    if sels is None or len(sels) == 0:
        if allow_no_selection:
            nr, nc = model.value.shape
            return (0, nr), (0, nc)
        raise ValueError("Table has no selection.")
    if len(sels) != 1:
        raise ValueError("Table must not contain multiple selections.")
    return sels[0]


def _to_clipboard_data_model(model: WidgetDataModel, format: str) -> ClipboardDataModel:
    (r0, r1), (c0, c1) = _get_selection(model, allow_no_selection=True)
    string, _, _ = table_to_text(model.value[r0:r1, c0:c1], format=format)
    return ClipboardDataModel(text=string)



from __future__ import annotations

from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtCore, QtGui
from superqt import QIconifyIcon

from himena.widgets import set_status_tip
from himena_builtins.qt.widgets._image_components._graphics_view import MouseMode
from himena_builtins.qt.widgets._image_components import _roi_items
from himena.qt._utils import qsignal_blocker

if TYPE_CHECKING:
    from himena_builtins.qt.widgets.image import QImageView


def _tool_btn(
    icon_name: str,
    tooltip: str,
    color: QtGui.QColor = QtGui.QColor(0, 0, 0),
) -> QtW.QToolButton:
    btn = QtW.QToolButton()
    btn.setIcon(_tool_btn_icon(icon_name, color=color.name()))
    btn.setCheckable(True)
    btn.setToolTip(tooltip)
    btn.setFixedSize(22, 22)
    return btn


class QRoiToolButton(QtW.QToolButton):
    def __init__(
        self,
        roi: _roi_items.QRoi,
        tooltip: str,
        color: QtGui.QColor = QtGui.QColor(0, 0, 0),
    ):
        super().__init__()
        self._roi = roi
        self.setIcon(_roi_tool_btn_icon(roi, color=color))
        self.setCheckable(True)
        self.setToolTip(tooltip)
        self.setFixedSize(22, 22)


def _roi_tool_btn_icon(roi: _roi_items.QRoi, color: QtGui.QColor):
    pixmap = QtGui.QPixmap(20, 20)
    pixmap.fill(QtCore.Qt.GlobalColor.transparent)
    pen = QtGui.QPen(color, 2)
    pen.setCosmetic(True)
    pen.setJoinStyle(QtCore.Qt.PenJoinStyle.MiterJoin)
    icon = QtGui.QIcon(roi.withPen(pen).makeThumbnail(pixmap))
    return icon


def _tool_btn_icon(icon_name: str, color: str) -> QtGui.QIcon:
    return QIconifyIcon(icon_name, color=color)


ICON_ZOOM = "mdi:magnify-expand"
ICON_SELECT = "mdi:cursor-default"

_THUMBNAIL_ROIS: dict[MouseMode, _roi_items.QRoi] = {
    MouseMode.ROI_RECTANGLE: _roi_items.QRectangleRoi(0, 0, 10, 8),
    MouseMode.ROI_ROTATED_RECTANGLE: _roi_items.QRotatedRectangleRoi(
        QtCore.QPointF(0, -2), QtCore.QPointF(4, 2), 3.6
    ),
    MouseMode.ROI_ELLIPSE: _roi_items.QEllipseRoi(0, 0, 10, 8),
    MouseMode.ROI_ROTATED_ELLIPSE: _roi_items.QRotatedEllipseRoi(
        QtCore.QPointF(0, -2), QtCore.QPointF(4, 2), 3.6
    ),
    MouseMode.ROI_CIRCLE: _roi_items.QCircleRoi(0, 0, 5),
    MouseMode.ROI_LINE: _roi_items.QLineRoi(0, 0, 10, 8),
    MouseMode.ROI_SEGMENTED_LINE: _roi_items.QSegmentedLineRoi(
        [0, 4, 8, 12], [10, 4, 6, 0]
    ),
    MouseMode.ROI_POLYGON: _roi_items.QPolygonRoi(
        [0, -5, -3, 3, 5, 0], [-2, -5, 3, 3, -5, -2]
    ),
    MouseMode.ROI_POINT: _roi_items.QPointRoi(0, 0),
    MouseMode.ROI_POINTS: _roi_items.QPointsRoi([], []),
}


class QRoiButtons(QtW.QWidget):
    def __init__(self, view: QImageView):
        super().__init__()
        self._img_view = view._img_view
        layout = QtW.QGridLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(1)
        layout.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignTop
        )
        self._btn_map: dict[MouseMode, QtW.QToolButton] = {
            MouseMode.PAN_ZOOM: _tool_btn(
                icon_name=ICON_ZOOM,
                tooltip="Pan/zoom mode (Z, Space)",
            ),
            MouseMode.SELECT: _tool_btn(
                icon_name=ICON_SELECT,
                tooltip="Select mode (S)",
            ),
            MouseMode.ROI_RECTANGLE: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_RECTANGLE],
                tooltip="Add rectangles (R)",
            ),
            MouseMode.ROI_ROTATED_RECTANGLE: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_ROTATED_RECTANGLE],
                tooltip="Add rotated rectangles (Shift+R)",
            ),
            MouseMode.ROI_ELLIPSE: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_ELLIPSE],
                tooltip="Add ellipses (E)",
            ),
            MouseMode.ROI_ROTATED_ELLIPSE: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_ROTATED_ELLIPSE],
                tooltip="Add rotated ellipses (Shift+E)",
            ),
            MouseMode.ROI_LINE: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_LINE],
                tooltip="Add lines (L)",
            ),
            MouseMode.ROI_SEGMENTED_LINE: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_SEGMENTED_LINE],
                tooltip="Add segmented lines (Shift+L)",
            ),
            MouseMode.ROI_POLYGON: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_POLYGON],
                tooltip="Add polygons (G)",
            ),
            MouseMode.ROI_CIRCLE: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_CIRCLE],
                tooltip="Add circles (C)",
            ),
            MouseMode.ROI_POINT: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_POINT],
                tooltip="Add points (P)",
            ),
            MouseMode.ROI_POINTS: QRoiToolButton(
                _THUMBNAIL_ROIS[MouseMode.ROI_POINTS],
                tooltip="Add multiple points (Shift+P)",
            ),
        }
        self._btn_map_inv = {v: k for k, v in self._btn_map.items()}
        self._button_group = QtW.QButtonGroup()
        for btn in self._btn_map.values():
            self._button_group.addButton(btn)
        self._button_group.setExclusive(True)
        self._button_group.buttonReleased.connect(self.btn_released)

        layout.addWidget(self._btn_map[MouseMode.PAN_ZOOM], 0, 0)
        layout.addWidget(self._btn_map[MouseMode.SELECT], 0, 1)
        layout.addWidget(self._btn_map[MouseMode.ROI_RECTANGLE], 1, 0)
        layout.addWidget(self._btn_map[MouseMode.ROI_ROTATED_RECTANGLE], 1, 1)
        layout.addWidget(self._btn_map[MouseMode.ROI_ELLIPSE], 1, 2)
        layout.addWidget(self._btn_map[MouseMode.ROI_ROTATED_ELLIPSE], 1, 3)
        layout.addWidget(self._btn_map[MouseMode.ROI_LINE], 2, 0)
        layout.addWidget(self._btn_map[MouseMode.ROI_SEGMENTED_LINE], 2, 1)
        layout.addWidget(self._btn_map[MouseMode.ROI_POLYGON], 2, 2)
        layout.addWidget(self._btn_map[MouseMode.ROI_CIRCLE], 2, 3)
        layout.addWidget(self._btn_map[MouseMode.ROI_POINT], 3, 0)
        layout.addWidget(self._btn_map[MouseMode.ROI_POINTS], 3, 1)

        self.setFixedHeight(84)
        self._btn_map[MouseMode.PAN_ZOOM].setChecked(True)
        self._img_view.mode_changed.connect(self.set_mode)

    def set_mode(self, mode: MouseMode):
        if btn := self._btn_map.get(mode):
            with qsignal_blocker(self._button_group):
                btn.setChecked(True)
        else:
            with qsignal_blocker(self._button_group):
                for button in self._button_group.buttons():
                    button.setChecked(False)

    def btn_released(self, btn: QtW.QToolButton):
        mode = self._btn_map_inv[btn]
        self._img_view.switch_mode(mode)
        mode_name = mode.name.replace("_", " ")
        if mode_name.startswith("ROI "):
            mode_name = mode_name[4:]
        set_status_tip(f"Switched to {mode_name} mode.")

    def _update_colors(self, color: QtGui.QColor):
        for mode, btn in self._btn_map.items():
            if roi := _THUMBNAIL_ROIS.get(mode):
                btn.setIcon(_roi_tool_btn_icon(roi, color))
        self._btn_map[MouseMode.PAN_ZOOM].setIcon(
            _tool_btn_icon(ICON_ZOOM, color=color.name())
        )
        self._btn_map[MouseMode.SELECT].setIcon(
            _tool_btn_icon(ICON_SELECT, color=color.name())
        )



from typing import TYPE_CHECKING, Iterator
import itertools

from cmap import Color, Colormap
import numpy as np

from himena._utils import to_color_or_colormap
from himena.plugins import register_function, configure_gui, configure_submenu
from himena.types import Parametric, WidgetDataModel
from himena.utils.table_selection import (
    model_to_xy_arrays,
    table_selection_gui_option,
    auto_select,
)
from himena.consts import StandardType, MenuId
from himena.widgets import SubWindow
from himena.standards import plotting as hplt
from himena.qt.magicgui import (
    FacePropertyEdit,
    EdgePropertyEdit,
    AxisPropertyEdit,
    DictEdit,
)
from himena._app_model import AppContext as _ctx

if TYPE_CHECKING:
    from himena.qt.magicgui._plot_elements import FacePropertyDict, EdgePropertyDict

_TABLE_LIKE = [
    StandardType.TABLE,
    StandardType.ARRAY,
    StandardType.DATAFRAME,
    StandardType.EXCEL,
]
_MENU = [MenuId.TOOLS_PLOT, "/model_menu/plot"]
_EDGE_ONLY_VALUE = {"color": "tab10", "width": 2.0}
_NOT_AN_IMAGE = _ctx.active_window_model_subtype_1 != "image"

# Single 2D selection in the form of ((row start, row stop), (col start, col stop))
# We should avoid using slice because it is not serializable.
SelectionType = tuple[tuple[int, int], tuple[int, int]]

configure_submenu(MenuId.TOOLS_PLOT, group="20_builtins", order=32)


@register_function(
    title="Scatter Plot ...",
    types=_TABLE_LIKE,
    menus=_MENU,
    command_id="builtins:plot:scatter",
    enablement=_NOT_AN_IMAGE,
)
def scatter_plot(win: SubWindow) -> Parametric:
    """Make a scatter plot from a table-like data."""
    x0, y0 = auto_select(win.to_model(), 2)

    @configure_gui(
        x=table_selection_gui_option(win, default=x0),
        y=table_selection_gui_option(win, default=y0),
        face={"widget_type": FacePropertyEdit},
        edge={"widget_type": EdgePropertyEdit},
    )
    def configure_plot(
        x: SelectionType | None,
        y: SelectionType,
        symbol: str = "o",
        size: float = 6.0,
        face: dict = {},
        edge: dict = {},
    ) -> WidgetDataModel:
        model = win.to_model()
        fig = hplt.figure()
        xarr, yarrs = _get_xy_data(model, x, y, fig.axes)
        for name_yarr, _face, _edge in zip(
            yarrs, _iter_face(face), _iter_edge(edge, prefix="edge_")
        ):
            name, yarr = name_yarr
            fig.axes.scatter(
                xarr, yarr, symbol=symbol, size=size, name=name, **_face, **_edge
            )
        if len(yarrs) == 1:
            fig.axes.y.label = name
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return configure_plot


@register_function(
    title="Line Plot ...",
    types=_TABLE_LIKE,
    menus=_MENU,
    command_id="builtins:plot:line",
    enablement=_NOT_AN_IMAGE,
)
def line_plot(win: SubWindow) -> Parametric:
    """Make a line plot from a table-like data."""
    x0, y0 = auto_select(win.to_model(), 2)

    @configure_gui(
        x=table_selection_gui_option(win, default=x0),
        y=table_selection_gui_option(win, default=y0),
        edge={"widget_type": EdgePropertyEdit, "value": _EDGE_ONLY_VALUE},
    )
    def configure_plot(
        x: SelectionType | None,
        y: SelectionType,
        edge: dict = {},
    ) -> WidgetDataModel:
        model = win.to_model()
        fig = hplt.figure()
        xarr, yarrs = _get_xy_data(model, x, y, fig.axes)
        for name_yarr, _edge in zip(yarrs, _iter_edge(edge)):
            name, yarr = name_yarr
            fig.axes.plot(xarr, yarr, name=name, **_edge)
        if len(yarrs) == 1:
            fig.axes.y.label = name
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return configure_plot


@register_function(
    title="Bar Plot ...",
    types=_TABLE_LIKE,
    menus=_MENU,
    command_id="builtins:plot:bar",
    enablement=_NOT_AN_IMAGE,
)
def bar_plot(win: SubWindow) -> Parametric:
    """Make a bar plot from a table-like data."""
    model = win.to_model()
    x0, y0 = auto_select(model, 2)

    @configure_gui(
        x=table_selection_gui_option(win, default=x0),
        y=table_selection_gui_option(win, default=y0),
        bottom=table_selection_gui_option(win),
        face={"widget_type": FacePropertyEdit},
        edge={"widget_type": EdgePropertyEdit},
    )
    def configure_plot(
        x: SelectionType | None,
        y: SelectionType,
        bottom: SelectionType | None = None,
        bar_width: float = 0.8,
        face: dict = {},
        edge: dict = {},
    ) -> WidgetDataModel:
        model = win.to_model()
        fig = hplt.figure()
        if bottom is not None:
            _, bottoms = _get_xy_data(model, x, bottom, fig.axes)
        else:
            bottoms = itertools.repeat(None)
        xarr, yarrs = _get_xy_data(model, x, y, fig.axes)
        for name_yarr, name_bottom, _face, _edge in zip(
            yarrs, bottoms, _iter_face(face), _iter_edge(edge, prefix="edge_")
        ):
            name, yarr = name_yarr
            fig.axes.bar(
                xarr, yarr, bottom=_ignore_label(name_bottom), bar_width=bar_width,
                name=name, **_face, **_edge
            )  # fmt: skip
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return configure_plot


@register_function(
    title="Errorbar Plot ...",
    types=_TABLE_LIKE,
    menus=_MENU,
    command_id="builtins:plot:errorbar",
    enablement=_NOT_AN_IMAGE,
)
def errorbar_plot(win: SubWindow) -> Parametric:
    """Make an error bar plot from a table-like data."""
    x0, y0 = auto_select(win.to_model(), 2)

    @configure_gui(
        x=table_selection_gui_option(win, default=x0),
        y=table_selection_gui_option(win, default=y0),
        xerr=table_selection_gui_option(win),
        yerr=table_selection_gui_option(win),
        edge={"widget_type": EdgePropertyEdit, "value": _EDGE_ONLY_VALUE},
    )
    def configure_plot(
        x: SelectionType | None,
        y: SelectionType,
        xerr: SelectionType | None,
        yerr: SelectionType | None,
        capsize: float = 0.0,
        edge: dict = {},
    ) -> WidgetDataModel:
        model = win.to_model()
        fig = hplt.figure()
        if xerr is not None:
            _, xerrs = _get_xy_data(model, x, xerr, fig.axes)
        else:
            xerrs = itertools.repeat(None)
        if yerr is not None:
            _, yerrs = _get_xy_data(model, x, yerr, fig.axes)
        else:
            yerrs = itertools.repeat(None)
        xarr, yarrs = _get_xy_data(model, x, y, fig.axes)
        for name_yarr, _xer, _yer, _edge in zip(yarrs, xerrs, yerrs, _iter_edge(edge)):
            name, yarr = name_yarr
            fig.axes.errorbar(
                xarr, yarr, x_error=_ignore_label(_xer), y_error=_ignore_label(_yer),
                capsize=capsize, name=name, **_edge,
            )  # fmt: skip
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return configure_plot


@register_function(
    title="Band Plot ...",
    types=_TABLE_LIKE,
    menus=_MENU,
    command_id="builtins:plot:band",
    enablement=_NOT_AN_IMAGE,
)
def band_plot(win: SubWindow) -> Parametric:
    """Make a band plot from a table-like data."""
    x0, y10, y20 = auto_select(win.to_model(), 3)

    @configure_gui(
        x=table_selection_gui_option(win, default=x0),
        y0=table_selection_gui_option(win, default=y10),
        y1=table_selection_gui_option(win, default=y20),
        face={"widget_type": FacePropertyEdit},
        edge={"widget_type": EdgePropertyEdit},
    )
    def configure_plot(
        x: SelectionType | None,
        y0: SelectionType,
        y1: SelectionType,
        face: dict = {},
        edge: dict = {},
    ) -> WidgetDataModel:
        model = win.to_model()
        fig = hplt.figure()
        xarr, ydata1 = _get_xy_data(model, x, y0, fig.axes)
        _, ydata2 = _get_xy_data(model, x, y1, fig.axes)
        _face = next(_iter_face(face), {})
        _edge = next(_iter_edge(edge, prefix="edge_"), {})
        if len(ydata1) == 1 and len(ydata2) == 1:
            name, yar1 = ydata1[0]
            _, yar2 = ydata2[0]
            fig.axes.band(xarr, yar1, yar2, name=name, **_face, **_edge)
        else:
            raise ValueError("Only one pair of y values is allowed.")
        fig.axes.y.label = name
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return configure_plot


@register_function(
    title="Histogram ...",
    types=_TABLE_LIKE,
    menus=_MENU,
    command_id="builtins:plot:histogram",
    enablement=_NOT_AN_IMAGE,
)
def histogram(win: SubWindow) -> Parametric:
    """Make a histogram from a table-like data."""
    x0 = auto_select(win.to_model(), 1)[0]
    assert x0 is not None  # when num == 1, it must be a tuple.
    row_sel = x0[0]
    ndata = row_sel[1] - row_sel[0]

    @configure_gui(
        x=table_selection_gui_option(win, default=x0),
        bins={"min": 1, "value": max(int(np.sqrt(ndata)), 2)},
        face={"widget_type": FacePropertyEdit},
        edge={"widget_type": EdgePropertyEdit},
    )
    def configure_plot(
        x: SelectionType | None,
        bins: int = 10,
        face: dict = {},
        edge: dict = {},
    ) -> WidgetDataModel:
        model = win.to_model()
        fig = hplt.figure()
        _, yarrs = _get_xy_data(model, None, x, fig.axes)
        for name_yarr, _face, _edge in zip(
            yarrs, _iter_face(face), _iter_edge(edge, prefix="edge_")
        ):
            name, yarr = name_yarr
            fig.axes.hist(yarr, bins=bins, name=name, **_face, **_edge)
        fig.axes.x.label = name
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return configure_plot


@register_function(
    title="Edit Plot ...",
    types=[StandardType.PLOT],
    menus=[MenuId.TOOLS_PLOT],
    command_id="builtins:plot:edit",
    keybindings="E",
)
def edit_plot(model: WidgetDataModel) -> Parametric:
    """Edit the appearance of the plot."""
    lo = _get_single_axes(model)
    plot_models = lo.axes.models
    gui_options = {
        "title": {"widget_type": "LineEdit", "value": lo.axes.title or ""},
        "x": {"widget_type": AxisPropertyEdit, "value": lo.axes.x.model_dump()},
        "y": {"widget_type": AxisPropertyEdit, "value": lo.axes.y.model_dump()},
    }
    for i, m in enumerate(plot_models):
        opt = {
            "label": f"#{i}",
            "widget_type": DictEdit,
            "options": m.plot_option_dict(),
            "value": m.model_dump(),
        }
        gui_options[f"element_{i}"] = opt

    @configure_gui(gui_options=gui_options)
    def run_edit_plot(
        title: str,
        x: dict,
        y: dict,
        **kwargs: dict,
    ) -> WidgetDataModel:
        lo.axes.title = title
        lo.axes.x = hplt.Axis.model_validate(x)
        lo.axes.y = hplt.Axis.model_validate(y)
        new_models = []
        for plot_model, value in zip(plot_models, kwargs.values()):
            dumped = plot_model.model_dump()
            dumped.update(value)
            new_models.append(plot_model.model_validate(dumped))
        lo.axes.models = new_models
        model.update_inplace = True
        return model

    return run_edit_plot


@register_function(
    title="3D Scatter Plot ...",
    types=_TABLE_LIKE,
    menus=_MENU,
    command_id="builtins:plot:scatter-plot-3d",
    group="plot-3d",
    enablement=_NOT_AN_IMAGE,
)
def scatter_plot_3d(win: SubWindow) -> Parametric:
    """3D scatter plot."""
    x0, y0, z0 = auto_select(win.to_model(), 3)

    @configure_gui(
        x=table_selection_gui_option(win, default=x0),
        y=table_selection_gui_option(win, default=y0),
        z=table_selection_gui_option(win, default=z0),
        face={"widget_type": FacePropertyEdit},
        edge={"widget_type": EdgePropertyEdit},
    )
    def configure_plot(
        x: SelectionType,
        y: SelectionType,
        z: SelectionType,
        symbol: str = "o",
        size: float = 6.0,
        face: dict = {},
        edge: dict = {},
    ) -> WidgetDataModel:
        model = win.to_model()
        fig = hplt.figure_3d()
        xarr, yarrs = _get_xy_data(model, x, y, fig.axes)
        _, zarrs = _get_xy_data(model, x, z, fig.axes)

        if len(yarrs) == 1 and len(zarrs) == 1:
            name_y, yarr = yarrs[0]
            name_z, zarr = zarrs[0]
            _face = next(_iter_face(face), {})
            _edge = next(_iter_edge(edge, prefix="edge_"), {})
            fig.axes.scatter(
                xarr, yarr, zarr, symbol=symbol, size=size, name=name_z, **_face,
                **_edge
            )  # fmt: skip
        else:
            raise ValueError("Only one pair of y values is allowed.")
        fig.axes.y.label = name_y
        fig.axes.z.label = name_z
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return configure_plot


@register_function(
    title="3D Line Plot ...",
    types=_TABLE_LIKE,
    menus=_MENU,
    command_id="builtins:plot:line-plot-3d",
    group="plot-3d",
    enablement=_NOT_AN_IMAGE,
)
def line_plot_3d(win: SubWindow) -> Parametric:
    """3D line plot."""
    x0, y0, z0 = auto_select(win.to_model(), 3)

    @configure_gui(
        x=table_selection_gui_option(win, default=x0),
        y=table_selection_gui_option(win, default=y0),
        z=table_selection_gui_option(win, default=z0),
        edge={"widget_type": EdgePropertyEdit, "value": _EDGE_ONLY_VALUE},
    )
    def configure_plot(
        x: SelectionType,
        y: SelectionType,
        z: SelectionType,
        edge: dict = {},
    ) -> WidgetDataModel:
        model = win.to_model()
        fig = hplt.figure_3d()
        xarr, yarrs = _get_xy_data(model, x, y, fig.axes)
        _, zarrs = _get_xy_data(model, x, z, fig.axes)
        if len(yarrs) == 1 and len(zarrs) == 1:
            name_y, yarr = yarrs[0]
            name_z, zarr = zarrs[0]
            _edge = next(_iter_edge(edge), {})
            fig.axes.plot(xarr, yarr, zarr, name=name_z, **_edge)
        else:
            raise ValueError("Only one pair of y values is allowed.")
        fig.axes.y.label = name_y
        fig.axes.z.label = name_z
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return configure_plot


@register_function(
    title="Plot to DataFrame ...",
    types=StandardType.PLOT,
    menus=[MenuId.TOOLS_PLOT],
    command_id="builtins:plot:plot-to-dataframe",
)
def plot_to_dataframe(model: WidgetDataModel) -> Parametric:
    """Convert a plot component to a DataFrame."""

    lo = _get_single_axes(model)
    plot_models = lo.axes.models

    @configure_gui(component={"choices": _make_choices(plot_models)})
    def run(component: int) -> WidgetDataModel:
        """Convert the selected plot component to a DataFrame."""
        plot_model = plot_models[component]
        if isinstance(plot_model, hplt.Histogram):
            df = {
                "height": plot_model.height,
                "bin_start": plot_model.bins[:-1],
                "bin_end": plot_model.bins[1:],
            }
        elif isinstance(plot_model, hplt.Texts):
            df = {"x": plot_model.x, "y": plot_model.y, "text": plot_model.texts}
        elif isinstance(plot_model, hplt.ErrorBar):
            df = {
                "x": plot_model.x,
                "y": plot_model.y,
                "x_error": plot_model.x_error,
                "y_error": plot_model.y_error,
            }
        elif isinstance(plot_model, hplt.Span):
            df = {"start": plot_model.start, "end": plot_model.end}
        elif isinstance(plot_model, hplt.models.PlotModelXY):
            df = {"x": plot_model.x, "y": plot_model.y}
        elif isinstance(plot_model, hplt.Band):
            df = {"x": plot_model.x, "y0": plot_model.y0, "y1": plot_model.y1}
        elif isinstance(plot_model, (hplt.Scatter3D, hplt.Line3D)):
            df = {"x": plot_model.x, "y": plot_model.y, "z": plot_model.z}
        else:
            raise NotImplementedError(f"Type {type(plot_model)} is not supported.")
        return WidgetDataModel(
            value=df,
            type=StandardType.DATAFRAME,
            title=f"Data of {model.title}",
        )

    return run


@register_function(
    title="Select Plot Components ...",
    types=StandardType.PLOT,
    menus=[MenuId.TOOLS_PLOT],
    command_id="builtins:plot:select-plot-components",
)
def select_plot_components(model: WidgetDataModel) -> Parametric:
    """Select a subset of the plot component."""
    lo = _get_single_axes(model)
    plot_models = lo.axes.models

    @configure_gui(
        components={"choices": _make_choices(plot_models), "widget_type": "Select"}
    )
    def run(components: list[int]) -> WidgetDataModel:
        models = [plot_models[i] for i in components]
        axes = lo.axes.model_copy(update={"models": models})
        out = lo.model_copy(update={"axes": axes})
        return model.with_value(out).with_title_numbering()

    return run


@register_function(
    title="Concatenate With ...",
    types=StandardType.PLOT,
    menus=[MenuId.TOOLS_PLOT],
    command_id="builtins:plot:concatenate-with",
)
def concatenate_with(model: WidgetDataModel) -> Parametric:
    """Concatenate the plot with another plot."""
    lo = _get_single_axes(model)

    @configure_gui(others={"types": StandardType.PLOT})
    def run(others: list[WidgetDataModel]) -> WidgetDataModel:
        """Concatenate the plot with another plot."""
        out = lo
        for other in others:
            lo_other = _get_single_axes(other)
            out = out.merge_with(lo_other)
        return model.with_value(out).with_title_numbering()

    return run


def _get_xy_data(
    model: WidgetDataModel,
    x: SelectionType | None,
    y: SelectionType | None,
    axes: hplt.Axes,
) -> "tuple[np.ndarray, list[tuple[str | None, np.ndarray]]]":
    xarr, yarrs = model_to_xy_arrays(model, x, y)
    if xarr.name:
        axes.x.label = xarr.name
    return xarr.array, yarrs


def _iter_face(face: "FacePropertyDict | dict", prefix: str = "") -> Iterator[dict]:
    color = to_color_or_colormap(face.get("color", "gray"))
    hatch = face.get("hatch", None)
    if isinstance(color, Colormap):
        cycler = itertools.cycle(color.color_stops.colors)
    else:
        cycler = itertools.repeat(Color(color))
    while True:
        yield {f"{prefix}color": next(cycler), f"{prefix}hatch": hatch}


def _iter_edge(edge: "EdgePropertyDict | dict", prefix: str = "") -> Iterator[dict]:
    color = to_color_or_colormap(edge.get("color", "gray"))
    width = edge.get("width", None)
    style = edge.get("style", None)
    if isinstance(color, Colormap):
        cycler = itertools.cycle(color.color_stops.colors)
    else:
        cycler = itertools.repeat(Color(color))
    while True:
        yield {
            f"{prefix}color": next(cycler),
            f"{prefix}width": width,
            f"{prefix}style": style,
        }


def _ignore_label(
    named_array: tuple[str | None, np.ndarray] | None,
) -> np.ndarray | None:
    if named_array is not None:
        _, val = named_array
    else:
        val = None
    return val


def _get_single_axes(model: WidgetDataModel) -> hplt.SingleAxes | hplt.SingleAxes3D:
    if not isinstance(lo := model.value, hplt.BaseLayoutModel):
        raise ValueError(f"Expected a layout model, got {type(lo)}")
    if not isinstance(lo, (hplt.SingleAxes, hplt.SingleAxes3D)):
        raise NotImplementedError("Only SingleAxes is supported for now.")
    return lo


def _make_choices(models: list[hplt.BasePlotModel]) -> list[tuple[str, int]]:
    return [(f"({i}) {m.name}", i) for i, m in enumerate(models)]



from __future__ import annotations
from typing import Any
import numpy as np
from himena.consts import StandardType
from himena.plugins import (
    ReproduceArgs,
    register_modification_tracker,
    register_hidden_function,
)
from himena.types import Parametric, WidgetDataModel

USE_DIFFLIB_LIMIT = 1000  # Limit for the number of characters to switch to difflib
USE_SPARSE_TABLE_DIFF_LIMIT = 100  # Limit for the size to switch to sparse table diff


@register_hidden_function(command_id="builtins:user-modification:text")
def reproduce_text_modification(model: WidgetDataModel) -> Parametric:
    if not isinstance(model.value, str):
        raise TypeError("Model value must be a string")

    def run(diff: list[tuple[str, int, int, str]]) -> WidgetDataModel:
        old_text = model.value
        new_text = old_text
        offset = 0
        for tag, i1, i2, new_part in diff:
            i1 += offset
            i2 += offset
            if tag == "!":
                new_text = new_text[:i1] + new_part + new_text[i2:]
                offset += len(new_part) - (i2 - i1)
            elif tag == "-":
                new_text = new_text[:i1] + new_text[i2:]
                offset -= i2 - i1
            elif tag == "+":
                new_text = new_text[:i1] + new_part + new_text[i1:]
                offset += len(new_part)
        model.value = new_text
        return model

    return run


@register_modification_tracker(type=StandardType.TEXT)
def text_modification_tracker(old: str, new: str) -> ReproduceArgs:
    """Track modifications to text widgets."""
    import difflib

    if len(new) < USE_DIFFLIB_LIMIT:
        if old.startswith(new):
            diff = [("-", 0, len(old), "")]
        elif new.startswith(old):
            diff = [("+", 0, len(old), new)]
        else:
            diff = [("!", 0, len(old), new)]
    else:
        # Get the diff between the old and new text
        diff: list[tuple[str, int, int, str]] = []
        for tag, i1, i2, j1, j2 in difflib.SequenceMatcher(
            None, old, new
        ).get_opcodes():
            if tag == "replace":
                diff.append(("!", i1, i2, new[j1:j2]))
            elif tag == "delete":
                diff.append(("-", i1, i2, ""))
            elif tag == "insert":
                diff.append(("+", i1, i2, new[j1:j2]))

    # Create the ReproduceArgs object
    return ReproduceArgs(
        command_id="builtins:user-modification:text",
        with_params={"diff": diff},
    )


@register_hidden_function(command_id="builtins:user-modification:table")
def reproduce_table_modification(model: WidgetDataModel) -> Parametric:
    def run(diff: dict[str, list[Any]]) -> WidgetDataModel:
        if "array" in diff:
            out = np.array(diff["array"], dtype=np.dtypes.StringDType())
        else:
            shape = diff["shape"]
            out = np.zeros(shape, dtype=np.dtypes.StringDType())
            r0, c0 = model.value.shape
            sl = slice(0, min(r0, shape[0])), slice(0, min(c0, shape[1]))
            out[sl] = model.value[sl]
            for row, col, value in zip(diff["rows"], diff["cols"], diff["values"]):
                out[row, col] = value
        model.value = out
        return model

    return run


@register_modification_tracker(type=StandardType.TABLE)
def table_modification_tracker(old: np.ndarray, new: np.ndarray) -> ReproduceArgs:
    """Track modifications to table widgets."""
    if new.size < USE_SPARSE_TABLE_DIFF_LIMIT or new.size > old.size * 2:
        diff = {"array": new.tolist()}
    else:
        old_normed = old.copy()
        if old.shape[0] > new.shape[0]:
            old_normed = old_normed[: new.shape[0]]
        elif old.shape[0] < new.shape[0]:
            r0, c0 = old_normed.shape
            old_normed = np.vstack(
                (old_normed, np.zeros((new.shape[0] - r0, c0), dtype=old_normed.dtype))
            )
        if old.shape[1] > new.shape[1]:
            old_normed = old_normed[:, : new.shape[1]]
        elif old.shape[1] < new.shape[1]:
            r0, c0 = old_normed.shape
            old_normed = np.hstack(
                (old_normed, np.zeros((r0, new.shape[1] - c0), dtype=old_normed.dtype))
            )

        rows, cols = np.where(old_normed != new)
        if len(rows) * 3 > new.size:
            diff = {"array": new.tolist()}
        else:
            diff = {
                "shape": new.shape,
                "rows": rows.tolist(),
                "cols": cols.tolist(),
                "values": new[rows, cols].tolist(),
            }
    # Create the ReproduceArgs object
    return ReproduceArgs(
        command_id="builtins:user-modification:table",
        with_params={"diff": diff},
    )



from __future__ import annotations

from typing import TYPE_CHECKING
import warnings
from qtpy import QtWidgets as QtW, QtCore

from magicgui import widgets as mgw
from magicgui.widgets.bases import ValuedContainerWidget
from psygnal import throttled
from himena.profile import AppProfile
from himena.plugins import AppActionRegistry
from himena.qt.magicgui import get_type_map, ToggleSwitch


if TYPE_CHECKING:
    from himena.widgets import MainWindow


class QPluginConfigs(QtW.QWidget):
    """Widget for editing plugin configs.

    All the built-in and user-defined plugins are listed here. Any dict-like objects,
    including dataclass and pydantic.BaseModel, can be used as the config and will be
    converted into a widget by magicgui.
    """

    def __init__(self, ui: MainWindow):
        super().__init__()
        outer_layout = QtW.QVBoxLayout(self)
        self._filter_edit = QPluginConfigFilter(self)
        self._filter_edit.setToolTip("Filter plugin config parameters by label text.")
        outer_layout.addWidget(self._filter_edit)
        self._scroll_area = QtW.QScrollArea()
        self._scroll_area.setWidgetResizable(True)
        outer_layout.addWidget(self._scroll_area)
        self._ui = ui
        _central_widget = QtW.QWidget()
        self._scroll_area.setWidget(_central_widget)
        layout = QtW.QVBoxLayout(_central_widget)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        layout.setContentsMargins(0, 0, 0, 0)
        self._plugin_id_to_containers: dict[str, mgw.Container[mgw.Widget]] = {}
        self._layout = layout
        self.add_config_forms()

    def add_config_forms(self):
        type_map = get_type_map()
        reg = AppActionRegistry.instance()
        with type_map.type_registered(bool, widget_type=LabeledToggleSwitch):
            for plugin_id, plugin_config in self._ui.app_profile.plugin_configs.items():
                try:
                    widgets: list[mgw.Widget] = []
                    plugin_title = reg._plugin_default_configs[plugin_id].title
                    plugin_config = plugin_config.copy()
                    for param, opt in plugin_config.items():
                        if not isinstance(opt, dict):
                            raise TypeError(f"Invalid config for {plugin_id}: {param}")
                        _opt = opt.copy()
                        value = _opt.pop("value")
                        annotation = _opt.pop("annotation", None)
                        widget_type = None
                        if "choices" in _opt:
                            annotation = None
                            widget_type = "ComboBox"
                        widget = type_map.create_widget(
                            value=value,
                            annotation=annotation,
                            widget_type=widget_type,
                            label=_make_label_text(plugin_title, param, _opt),
                            options=_opt,
                            name=param,
                        )
                        if isinstance(widget, mgw.EmptyWidget):
                            warnings.warn(
                                f"Plugin config of {plugin_id!r} returned an empty widget.",
                                UserWarning,
                                stacklevel=2,
                            )
                        widgets.append(widget)
                    container = mgw.Container(widgets=widgets, name=plugin_id)
                    self._plugin_id_to_containers[plugin_id] = container
                    container.changed.connect(self._update_configs)
                    self._layout.addWidget(container.native)
                except Exception as e:
                    warnings.warn(f"Failed to create config for {plugin_id}: {e}")
        self._layout.addWidget(QtW.QWidget(), 1)  # spacer

    def _update_configs(self, container: mgw.Container):
        return _update_configs_throttled(self._ui.app_profile, container)


class QPluginConfigFilter(QtW.QLineEdit):
    def __init__(self, parent: QPluginConfigs):
        super().__init__(parent)
        self.setPlaceholderText("Type to filter ...")
        self.textChanged.connect(self._filter_configs)

    def parent(self) -> QPluginConfigs:
        return super().parent()

    def _filter_configs(self, text: str):
        return _filter_configs(self.parent(), text)


@throttled(timeout=100)
def _update_configs_throttled(prof: AppProfile, container: mgw.Container):
    prof.update_plugin_config(container.name, **container.asdict())


@throttled(timeout=100)
def _filter_configs(self: QPluginConfigs, text: str):
    text = text.strip().lower()
    empty = text == ""
    for container in self._plugin_id_to_containers.values():
        at_least_one_visible = False
        for child in container:
            vis = empty or text in _remove_html_from_label_text(child.label).lower()
            child.visible = vis
            at_least_one_visible = at_least_one_visible or vis
        container.visible = at_least_one_visible


_LEFT = '<b><font color="#808080">'
_RIGHT = "</font></b>"


def _make_label_text(plugin_title: str, param: str, opt: dict) -> str:
    param_text = opt.pop("label", param.replace("_", " ").capitalize())
    return f"{_LEFT}{plugin_title}:{_RIGHT} {param_text}"


def _remove_html_from_label_text(text: str):
    nchars = len(_LEFT)
    return "".join(text[nchars:].split(_RIGHT, 1))


class LabeledToggleSwitch(ValuedContainerWidget):
    def __init__(self, value=True, **kwargs):
        self._inner = ToggleSwitch(value=value, label="")
        widgets = [self._inner]
        kwargs["widgets"] = widgets
        super().__init__(value, **kwargs)
        self.margins = (0, 0, 0, 0)
        self._inner.changed.connect(self.changed.emit)

    def get_value(self):
        return self._inner.get_value()

    def set_value(self, value):
        return self._inner.set_value(value)
